from __future__ import annotations

import logging
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Optional,
    Self,
    Sequence,
    cast,
)
from uuid import UUID

import sqlalchemy as sa
from sqlalchemy.dialects import postgresql as pgsql
from sqlalchemy.ext.asyncio import AsyncEngine as SAEngine
from sqlalchemy.ext.asyncio import AsyncSession as SASession
from sqlalchemy.orm import foreign, joinedload, relationship, selectinload
from sqlalchemy.types import VARCHAR, TypeDecorator

from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.auth.hash import PasswordHashAlgorithm
from ai.backend.manager.data.model_serving.types import UserData as ModelServingUserData
from ai.backend.manager.data.user.types import UserCreator, UserData, UserRole, UserStatus
from ai.backend.manager.models.hasher.types import HashInfo, PasswordInfo

from .base import (
    Base,
    EnumValueType,
    IDColumn,
    IPColumn,
    mapper_registry,
)
from .exceptions import ObjectNotFound
from .hasher import PasswordHasherFactory
from .types import (
    QueryCondition,
    QueryOption,
    load_related_field,
)
from .utils import ExtendedAsyncSAEngine, execute_with_txn_retry

if TYPE_CHECKING:
    from .keypair import KeyPairRow

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

__all__: Sequence[str] = (
    "users",
    "UserRow",
    "UserRole",  # For compatibility with existing code
    "UserStatus",  # For compatibility with existing code
    "ACTIVE_USER_STATUSES",
    "INACTIVE_USER_STATUSES",
    "PasswordHashAlgorithm",
    "compare_to_hashed_password",
    "check_credential_with_migration",
    "check_credential",
)


class PasswordColumn(TypeDecorator):
    """Custom column type that prevents direct password assignment.

    Passwords should be set using proper functions that have access to config:
    - Use check_credential() for login with gradual migration
    - Use explicit UPDATE queries with pre-hashed passwords

    This column type is kept for backward compatibility but should not be used
    for setting passwords directly.
    """

    impl = VARCHAR
    cache_ok = True

    def process_bind_param(self, value: Any, dialect: Any) -> Optional[str]:
        if value is None:
            return None

        if not isinstance(value, PasswordInfo):
            raise ValueError("Password must be set using PasswordInfo for hashing.")

        hash_info = value.generate_new_hash()
        return hash_info.to_string()


ACTIVE_USER_STATUSES = (UserStatus.ACTIVE,)

INACTIVE_USER_STATUSES = (
    UserStatus.INACTIVE,
    UserStatus.DELETED,
    UserStatus.BEFORE_VERIFICATION,
)


users = sa.Table(
    "users",
    mapper_registry.metadata,
    IDColumn("uuid"),
    sa.Column("username", sa.String(length=64), unique=True),
    sa.Column("email", sa.String(length=64), index=True, nullable=False, unique=True),
    sa.Column("password", PasswordColumn()),
    sa.Column("need_password_change", sa.Boolean),
    sa.Column(
        "password_changed_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
    ),
    sa.Column("full_name", sa.String(length=64)),
    sa.Column("description", sa.String(length=500)),
    sa.Column("status", EnumValueType(UserStatus), default=UserStatus.ACTIVE, nullable=False),
    sa.Column("status_info", sa.Unicode(), nullable=True, default=sa.null()),
    sa.Column("created_at", sa.DateTime(timezone=True), server_default=sa.func.now()),
    sa.Column(
        "modified_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        onupdate=sa.func.current_timestamp(),
    ),
    #: Field for synchronization with external services.
    sa.Column("integration_id", sa.String(length=512)),
    sa.Column("domain_name", sa.String(length=64), sa.ForeignKey("domains.name"), index=True),
    sa.Column("role", EnumValueType(UserRole), default=UserRole.USER),
    sa.Column("allowed_client_ip", pgsql.ARRAY(IPColumn), nullable=True),
    sa.Column("totp_key", sa.String(length=32)),
    sa.Column("totp_activated", sa.Boolean, server_default=sa.false(), default=False),
    sa.Column("totp_activated_at", sa.DateTime(timezone=True), nullable=True),
    sa.Column(
        "resource_policy",
        sa.String(length=256),
        sa.ForeignKey("user_resource_policies.name"),
        nullable=False,
    ),
    sa.Column(
        "sudo_session_enabled",
        sa.Boolean,
        default=False,
        nullable=False,
    ),
    sa.Column(
        "main_access_key",
        sa.String(length=20),
        sa.ForeignKey("keypairs.access_key", ondelete="SET NULL"),
        nullable=True,  # keypairs.user is non-nullable
    ),
    sa.Column("container_uid", sa.Integer, nullable=True, server_default=sa.null()),
    sa.Column("container_main_gid", sa.Integer, nullable=True, server_default=sa.null()),
    sa.Column("container_gids", sa.ARRAY(sa.Integer), nullable=True, server_default=sa.null()),
)


# Defined for avoiding circular import
def _get_session_row_join_condition():
    from ai.backend.manager.models.session import SessionRow

    return UserRow.uuid == foreign(SessionRow.user_uuid)


def _get_kernel_row_join_condition():
    from ai.backend.manager.models.kernel import KernelRow

    return UserRow.uuid == foreign(KernelRow.user_uuid)


class UserRow(Base):
    __table__ = users
    # from .keypair import KeyPairRow

    sessions = relationship(
        "SessionRow",
        back_populates="user",
        primaryjoin=_get_session_row_join_condition,
        foreign_keys="SessionRow.user_uuid",
    )
    kernels = relationship(
        "KernelRow",
        back_populates="user_row",
        primaryjoin=_get_kernel_row_join_condition,
        foreign_keys="KernelRow.user_uuid",
    )
    domain = relationship("DomainRow", back_populates="users")
    groups = relationship("AssocGroupUserRow", back_populates="user")
    resource_policy_row = relationship("UserResourcePolicyRow", back_populates="users")
    keypairs = relationship("KeyPairRow", back_populates="user_row", foreign_keys="KeyPairRow.user")

    created_endpoints = relationship(
        "EndpointRow",
        back_populates="created_user_row",
        primaryjoin="foreign(EndpointRow.created_user) == UserRow.uuid",
    )
    owned_endpoints = relationship(
        "EndpointRow",
        back_populates="session_owner_row",
        primaryjoin="foreign(EndpointRow.session_owner) == UserRow.uuid",
    )

    main_keypair = relationship("KeyPairRow", foreign_keys=users.c.main_access_key)

    vfolder_rows = relationship(
        "VFolderRow",
        back_populates="user_row",
        primaryjoin="UserRow.uuid == foreign(VFolderRow.user)",
    )

    role_assignments = relationship(
        "UserRoleRow",
        back_populates="user_row",
        primaryjoin="UserRow.uuid == foreign(UserRoleRow.user_id)",
    )

    @classmethod
    def from_creator(cls, creator: UserCreator) -> Self:
        return cls(
            username=creator.username,
            email=creator.email,
            password=creator.password,
            need_password_change=creator.need_password_change
            if creator.need_password_change is not None
            else False,
            full_name=creator.full_name,
            description=creator.description,
            status=creator.status if creator.status is not None else UserStatus.BEFORE_VERIFICATION,
            domain_name=creator.domain_name,
            role=creator.role if creator.role is not None else UserRole.USER,
            resource_policy=creator.resource_policy
            if creator.resource_policy is not None
            else "default",
            allowed_client_ip=creator.allowed_client_ip,
            totp_activated=creator.totp_activated if creator.totp_activated is not None else False,
            sudo_session_enabled=creator.sudo_session_enabled
            if creator.sudo_session_enabled is not None
            else False,
            container_uid=creator.container_uid,
            container_main_gid=creator.container_main_gid,
            container_gids=creator.container_gids,
        )

    @classmethod
    def load_keypairs(cls) -> Callable:
        from .keypair import KeyPairRow

        return selectinload(UserRow.keypairs).options(joinedload(KeyPairRow.resource_policy_row))

    @classmethod
    def load_main_keypair(cls) -> Callable:
        from .keypair import KeyPairRow

        return joinedload(UserRow.main_keypair).options(joinedload(KeyPairRow.resource_policy_row))

    @classmethod
    def load_resource_policy(cls) -> Callable:
        return joinedload(UserRow.resource_policy_row)

    @classmethod
    async def query_user_by_uuid(
        cls,
        user_uuid: UUID,
        db_session: SASession,
    ) -> Optional[Self]:
        user_query = (
            sa.select(UserRow)
            .where(UserRow.uuid == user_uuid)
            .options(
                joinedload(UserRow.main_keypair),
                selectinload(UserRow.keypairs),
            )
        )
        user_row = await db_session.scalar(user_query)
        return user_row

    @classmethod
    async def query_by_condition(
        cls,
        conditions: Sequence[QueryCondition],
        options: Sequence[QueryOption] = tuple(),
        *,
        db: ExtendedAsyncSAEngine,
    ) -> list[Self]:
        """
        Query user rows by condition.
        Args:
            condition: QueryCondition.
            options: A sequence of query options.
            db: Database engine.
        Returns:
            A list of UserRow instances that match the condition.
        Raises:
            EmptySQLCondition: If condition is empty.
        """
        query_stmt = sa.select(UserRow)
        for cond in conditions:
            query_stmt = cond(query_stmt)

        for option in options:
            query_stmt = option(query_stmt)

        async def fetch(db_session: SASession) -> list[Self]:
            return (await db_session.scalars(query_stmt)).all()

        async with db.connect() as db_conn:
            return await execute_with_txn_retry(
                fetch,
                db.begin_readonly_session,
                db_conn,
            )

    @classmethod
    async def get_by_id_with_policies(
        cls,
        user_uuid: UUID,
        *,
        db: ExtendedAsyncSAEngine,
    ) -> Self:
        """
        Query user row by UUID with related policies.
        Args:
            user_uuid: User UUID.
            db: Database engine.
        Returns:
            The UserRow instance that matches the UUID.
        Raises:
            ObjectNotFound: If user not found.
        """
        rows = await cls.query_by_condition(
            [by_user_uuid(user_uuid)],
            [
                load_related_field(cls.load_keypairs),
                load_related_field(cls.load_main_keypair),
                load_related_field(cls.load_resource_policy),
            ],
            db=db,
        )
        if not rows:
            raise ObjectNotFound(f"User with id {user_uuid} not found")
        return rows[0]

    def get_main_keypair_row(self) -> Optional[KeyPairRow]:
        # `cast()` requires import of KeyPairRow
        from .keypair import KeyPairRow

        keypair_candidate: Optional[KeyPairRow] = None
        main_keypair_row = cast(Optional[KeyPairRow], self.main_keypair)
        if main_keypair_row is None:
            keypair_rows = cast(list[KeyPairRow], self.keypairs)
            active_keypairs = [row for row in keypair_rows if row.is_active]
            for row in active_keypairs:
                if keypair_candidate is None or not keypair_candidate.is_admin:
                    keypair_candidate = row
                    break
            if keypair_candidate is not None:
                self.main_keypair = keypair_candidate
        else:
            keypair_candidate = main_keypair_row
        return keypair_candidate

    def to_model_serving_user_data(self) -> ModelServingUserData:
        return ModelServingUserData(
            uuid=self.uuid,
            email=self.email,
        )

    def to_data(self) -> UserData:
        return UserData(
            id=self.uuid,
            uuid=self.uuid,
            username=self.username,
            email=self.email,
            need_password_change=self.need_password_change,
            full_name=self.full_name,
            description=self.description,
            is_active=self.status == UserStatus.ACTIVE,
            status=self.status.value,
            status_info=self.status_info,
            created_at=self.created_at,
            modified_at=self.modified_at,
            domain_name=self.domain_name,
            role=self.role.value,
            resource_policy=self.resource_policy,
            allowed_client_ip=self.allowed_client_ip,
            totp_activated=self.totp_activated,
            totp_activated_at=self.totp_activated_at,
            sudo_session_enabled=self.sudo_session_enabled,
            main_access_key=self.main_access_key,
            container_uid=self.container_uid,
            container_main_gid=self.container_main_gid,
            container_gids=self.container_gids,
        )


def by_user_uuid(
    user_uuid: UUID,
) -> QueryCondition:
    def _by_user_uuid(
        query_stmt: sa.sql.Select,
    ) -> sa.sql.Select:
        return query_stmt.where(UserRow.uuid == user_uuid)

    return _by_user_uuid


def by_username(
    username: str,
) -> QueryCondition:
    def _by_username(
        query_stmt: sa.sql.Select,
    ) -> sa.sql.Select:
        return query_stmt.where(UserRow.username == username)

    return _by_username


def by_user_email(
    email: str,
) -> QueryCondition:
    def _by_email(
        query_stmt: sa.sql.Select,
    ) -> sa.sql.Select:
        return query_stmt.where(UserRow.email == email)

    return _by_email


def _verify_password(guess: str, hashed: str) -> bool:
    """Verify a password against a hashed password."""
    hash_info = HashInfo.from_hash_string(hashed)
    hasher = PasswordHasherFactory.get_hasher(hash_info.algorithm)
    return hasher.verify(guess, hash_info)


def compare_to_hashed_password(raw_password: str, hashed_password: str) -> bool:
    """
    Compare a raw string password value to hashed password.
    """
    return _verify_password(raw_password, hashed_password)


async def check_credential_with_migration(
    db: SAEngine,
    domain: str,
    email: str,
    target_password_info: PasswordInfo,
) -> Any:
    """
    Check user credentials and optionally migrate password hash if needed.

    Args:
        db: Database engine
        domain: User's domain
        email: User's email
        target_password_info: Password configuration containing password and target hash settings

    Returns:
        User row if credentials are valid, None otherwise
    """
    async with db.begin_readonly() as conn:
        result = await conn.execute(
            sa.select([users])
            .select_from(users)
            .where(
                (users.c.email == email) & (users.c.domain_name == domain),
            ),
        )
    row = result.first()
    if row is None:
        return None
    if row["password"] is None:
        # user password is not set.
        return None

    try:
        if not _verify_password(target_password_info.password, row["password"]):
            return None
    except ValueError:
        return None

    # Password is valid, check if we need to migrate the hash
    current_hash_info = HashInfo.from_hash_string(row["password"])
    if current_hash_info is None:
        # Shouldn't happen since password was just verified
        return row

    if target_password_info.need_migration(current_hash_info):
        # Re-hash the password with the new algorithm using the provided PasswordInfo
        # Update the user's password hash asynchronously
        async with db.begin() as conn:
            await conn.execute(
                sa.update(users)
                .where((users.c.email == email) & (users.c.domain_name == domain))
                .values(password=target_password_info)
            )

    return row


async def check_credential(
    db: SAEngine,
    domain: str,
    email: str,
    password: str,
) -> Any:
    """
    Check user credentials without migration (for signout, update password, etc.)

    Args:
        db: Database engine
        domain: User's domain
        email: User's email
        password: Plain text password to verify

    Returns:
        User row if credentials are valid, None otherwise
    """
    async with db.begin_readonly() as conn:
        result = await conn.execute(
            sa.select([users])
            .select_from(users)
            .where(
                (users.c.email == email) & (users.c.domain_name == domain),
            ),
        )
    row = result.first()
    if row is None:
        return None
    if row["password"] is None:
        # user password is not set.
        return None

    try:
        if not _verify_password(password, row["password"]):
            return None
    except ValueError:
        return None

    return row



import uuid
from typing import Optional

from ai.backend.common.data.artifact.types import VerificationStepResult
from ai.backend.common.data.storage.registries.types import ModelData
from ai.backend.common.data.storage.types import ArtifactStorageType
from ai.backend.common.exception import BackendAIError
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience
from ai.backend.manager.data.artifact.modifier import ArtifactModifier
from ai.backend.manager.data.artifact.types import (
    ArtifactData,
    ArtifactDataWithRevisions,
    ArtifactFilterOptions,
    ArtifactOrderingOptions,
    ArtifactRemoteStatus,
    ArtifactRevisionData,
    ArtifactStatus,
)
from ai.backend.manager.data.association.types import AssociationArtifactsStoragesData
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.repositories.artifact.db_source.db_source import ArtifactDBSource
from ai.backend.manager.repositories.artifact.types import (
    ArtifactRevisionFilterOptions,
    ArtifactRevisionOrderingOptions,
)
from ai.backend.manager.repositories.types import PaginationOptions

artifact_repository_resilience = Resilience(
    policies=[
        MetricPolicy(MetricArgs(domain=DomainType.REPOSITORY, layer=LayerType.ARTIFACT_REPOSITORY)),
        RetryPolicy(
            RetryArgs(
                max_retries=10,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class ArtifactRepository:
    """Repository layer that delegates to data source."""

    _db_source: ArtifactDBSource

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db_source = ArtifactDBSource(db)

    @artifact_repository_resilience.apply()
    async def get_artifact_by_id(self, artifact_id: uuid.UUID) -> ArtifactData:
        return await self._db_source.get_artifact_by_id(artifact_id)

    @artifact_repository_resilience.apply()
    async def get_artifact_revision_by_id(self, revision_id: uuid.UUID) -> ArtifactRevisionData:
        return await self._db_source.get_artifact_revision_by_id(revision_id)

    @artifact_repository_resilience.apply()
    async def get_model_artifact(self, model_id: str, registry_id: uuid.UUID) -> ArtifactData:
        return await self._db_source.get_model_artifact(model_id, registry_id)

    @artifact_repository_resilience.apply()
    async def get_artifact_revision(
        self, artifact_id: uuid.UUID, revision: str
    ) -> ArtifactRevisionData:
        return await self._db_source.get_artifact_revision(artifact_id, revision)

    @artifact_repository_resilience.apply()
    async def update_artifact(
        self, artifact_id: uuid.UUID, modifier: ArtifactModifier
    ) -> ArtifactData:
        return await self._db_source.update_artifact(artifact_id, modifier)

    @artifact_repository_resilience.apply()
    async def list_artifact_revisions(self, artifact_id: uuid.UUID) -> list[ArtifactRevisionData]:
        return await self._db_source.list_artifact_revisions(artifact_id)

    @artifact_repository_resilience.apply()
    async def upsert_artifacts(
        self,
        artifacts: list[ArtifactData],
    ) -> list[ArtifactData]:
        return await self._db_source.upsert_artifacts(artifacts)

    @artifact_repository_resilience.apply()
    async def upsert_artifact_revisions(
        self,
        revisions: list[ArtifactRevisionData],
    ) -> list[ArtifactRevisionData]:
        return await self._db_source.upsert_artifact_revisions(revisions)

    @artifact_repository_resilience.apply()
    async def upsert_huggingface_model_artifacts(
        self,
        model_list: list[ModelData],
        registry_id: uuid.UUID,
    ) -> list[ArtifactDataWithRevisions]:
        return await self._db_source.upsert_huggingface_model_artifacts(model_list, registry_id)

    @artifact_repository_resilience.apply()
    async def associate_artifact_with_storage(
        self,
        artifact_revision_id: uuid.UUID,
        storage_namespace_id: uuid.UUID,
        storage_type: ArtifactStorageType,
    ) -> AssociationArtifactsStoragesData:
        return await self._db_source.associate_artifact_with_storage(
            artifact_revision_id, storage_namespace_id, storage_type
        )

    @artifact_repository_resilience.apply()
    async def disassociate_artifact_with_storage(
        self, artifact_revision_id: uuid.UUID, storage_namespace_id: uuid.UUID
    ) -> AssociationArtifactsStoragesData:
        return await self._db_source.disassociate_artifact_with_storage(
            artifact_revision_id, storage_namespace_id
        )

    @artifact_repository_resilience.apply()
    async def approve_artifact(self, revision_id: uuid.UUID) -> ArtifactRevisionData:
        return await self._db_source.approve_artifact(revision_id)

    @artifact_repository_resilience.apply()
    async def reject_artifact(self, revision_id: uuid.UUID) -> ArtifactRevisionData:
        return await self._db_source.reject_artifact(revision_id)

    @artifact_repository_resilience.apply()
    async def reset_artifact_revision_status(self, revision_id: uuid.UUID) -> uuid.UUID:
        return await self._db_source.reset_artifact_revision_status(revision_id)

    @artifact_repository_resilience.apply()
    async def update_artifact_revision_status(
        self, artifact_revision_id: uuid.UUID, status: ArtifactStatus
    ) -> uuid.UUID:
        return await self._db_source.update_artifact_revision_status(artifact_revision_id, status)

    @artifact_repository_resilience.apply()
    async def update_artifact_revision_remote_status(
        self, artifact_revision_id: uuid.UUID, remote_status: ArtifactRemoteStatus
    ) -> uuid.UUID:
        return await self._db_source.update_artifact_revision_remote_status(
            artifact_revision_id, remote_status
        )

    @artifact_repository_resilience.apply()
    async def delete_artifacts(self, artifact_ids: list[uuid.UUID]) -> list[ArtifactData]:
        return await self._db_source.delete_artifacts(artifact_ids)

    @artifact_repository_resilience.apply()
    async def restore_artifacts(self, artifact_ids: list[uuid.UUID]) -> list[ArtifactData]:
        return await self._db_source.restore_artifacts(artifact_ids)

    @artifact_repository_resilience.apply()
    async def update_artifact_revision_bytesize(
        self, artifact_revision_id: uuid.UUID, size: int
    ) -> uuid.UUID:
        return await self._db_source.update_artifact_revision_bytesize(artifact_revision_id, size)

    @artifact_repository_resilience.apply()
    async def update_artifact_revision_digest(
        self, artifact_revision_id: uuid.UUID, digest: str
    ) -> uuid.UUID:
        return await self._db_source.update_artifact_revision_digest(artifact_revision_id, digest)

    @artifact_repository_resilience.apply()
    async def update_artifact_revision_readme(
        self, artifact_revision_id: uuid.UUID, readme: str
    ) -> uuid.UUID:
        return await self._db_source.update_artifact_revision_readme(artifact_revision_id, readme)

    @artifact_repository_resilience.apply()
    async def update_artifact_revision_verification_result(
        self, artifact_revision_id: uuid.UUID, verification_result: VerificationStepResult
    ) -> uuid.UUID:
        return await self._db_source.update_artifact_revision_verification_result(
            artifact_revision_id, verification_result.model_dump()
        )

    @artifact_repository_resilience.apply()
    async def get_artifact_revision_readme(self, artifact_revision_id: uuid.UUID) -> str:
        return await self._db_source.get_artifact_revision_readme(artifact_revision_id)

    @artifact_repository_resilience.apply()
    async def list_artifacts_paginated(
        self,
        *,
        pagination: Optional[PaginationOptions] = None,
        ordering: Optional[ArtifactOrderingOptions] = None,
        filters: Optional[ArtifactFilterOptions] = None,
    ) -> tuple[list[ArtifactData], int]:
        return await self._db_source.list_artifacts_paginated(
            pagination=pagination, ordering=ordering, filters=filters
        )

    @artifact_repository_resilience.apply()
    async def list_artifacts_with_revisions_paginated(
        self,
        *,
        pagination: Optional[PaginationOptions] = None,
        ordering: Optional[ArtifactOrderingOptions] = None,
        filters: Optional[ArtifactFilterOptions] = None,
    ) -> tuple[list[ArtifactDataWithRevisions], int]:
        return await self._db_source.list_artifacts_with_revisions_paginated(
            pagination=pagination, ordering=ordering, filters=filters
        )

    @artifact_repository_resilience.apply()
    async def list_artifact_revisions_paginated(
        self,
        *,
        pagination: Optional[PaginationOptions] = None,
        ordering: Optional[ArtifactRevisionOrderingOptions] = None,
        filters: Optional[ArtifactRevisionFilterOptions] = None,
    ) -> tuple[list[ArtifactRevisionData], int]:
        return await self._db_source.list_artifact_revisions_paginated(
            pagination=pagination, ordering=ordering, filters=filters
        )



import logging
from collections.abc import Mapping
from typing import Optional, cast
from uuid import UUID

import sqlalchemy as sa
from sqlalchemy.exc import DBAPIError
from sqlalchemy.ext.asyncio import AsyncSession as SASession
from sqlalchemy.orm import selectinload

from ai.backend.common.docker import ImageRef
from ai.backend.common.exception import UnknownImageReference
from ai.backend.common.types import ImageAlias, ImageID
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.data.image.types import (
    ImageAliasData,
    ImageData,
    ImageDataWithDetails,
    ImageStatus,
    RescanImagesResult,
)
from ai.backend.manager.errors.image import (
    AliasImageActionDBError,
    AliasImageActionValueError,
    ForgetImageForbiddenError,
    ImageAliasNotFound,
    ImageNotFound,
    ModifyImageActionValueError,
    PurgeImageActionByIdObjectDBError,
    RegistryNotFoundForImage,
)
from ai.backend.manager.models.container_registry import ContainerRegistryRow
from ai.backend.manager.models.image import (
    ImageAliasRow,
    ImageIdentifier,
    ImageRow,
    scan_single_image,
)
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ImageDBSource:
    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    async def fetch_image_by_identifiers(
        self, identifiers: list[ImageAlias | ImageRef | ImageIdentifier]
    ) -> ImageData:
        """
        Fetches an image from database by its identifiers, which can be a combination of
        ImageAlias, ImageRef, or ImageIdentifier.
        Returns an ImageData object.
        Raises Exception if the image cannot be found.
        """
        async with self._db.begin_session() as session:
            row = await self._resolve_image(session, identifiers)
            data = row.to_dataclass()
        return data

    async def fetch_images_batch(
        self, identifier_lists: list[list[ImageIdentifier]]
    ) -> list[ImageData]:
        """
        Fetches multiple images from database by their identifiers in a single database session.
        Returns a list of ImageData objects.
        More efficient than multiple individual fetch operations.
        """
        async with self._db.begin_session() as session:
            rows: list[ImageRow] = []
            for identifiers in identifier_lists:
                row = await self._resolve_image(
                    session, cast(list[ImageAlias | ImageRef | ImageIdentifier], identifiers)
                )
                rows.append(row)
            data_list = [row.to_dataclass() for row in rows]
        return data_list

    async def _resolve_image(
        self,
        session: SASession,
        identifiers: list[ImageAlias | ImageRef | ImageIdentifier],
    ) -> ImageRow:
        return await ImageRow.resolve(session, identifiers)

    async def _get_image_by_id(
        self,
        session: SASession,
        image_id: UUID,
        load_aliases: bool = False,
        status_filter: Optional[list[ImageStatus]] = None,
    ) -> ImageRow:
        """
        Private method to get an image by ID using an existing session.
        Returns None if image is not found.
        """
        row = await ImageRow.get(
            session, image_id, load_aliases=load_aliases, filter_by_statuses=status_filter
        )
        if row is None:
            raise ImageNotFound()
        return row

    async def _validate_image_ownership(
        self, session: SASession, image_id: UUID, user_id: UUID, load_aliases: bool = False
    ) -> ImageRow:
        """
        Private method to get an image and validate ownership using an existing session.
        Raises ForgetImageActionGenericForbiddenError if image doesn't exist or user doesn't own it.
        """
        image_row = await self._get_image_by_id(session, image_id, load_aliases)
        if not image_row.is_owned_by(user_id):
            raise ForgetImageForbiddenError()
        return image_row

    async def _get_image_alias_by_name(self, session: SASession, alias: str) -> ImageAliasRow:
        """
        Private method to get an image alias by name using an existing session.
        """
        image_alias_row = await session.scalar(
            sa.select(ImageAliasRow).where(ImageAliasRow.alias == alias),
        )
        if not image_alias_row:
            raise ImageAliasNotFound(f"Image alias '{alias}' not found.")
        return image_alias_row

    async def query_images_by_canonicals(
        self,
        canonicals: list[str],
        status_filter: Optional[list[ImageStatus]] = None,
    ) -> dict[ImageID, ImageDataWithDetails]:
        query = (
            sa.select(ImageRow)
            .where(ImageRow.name.in_(canonicals))
            .options(selectinload(ImageRow.aliases))
        )
        if status_filter:
            query = query.where(ImageRow.status.in_(status_filter))

        async with self._db.begin_readonly_session() as session:
            result = await session.execute(query)
            image_rows: list[ImageRow] = result.scalars().all()
            return {ImageID(row.id): row.to_detailed_dataclass() for row in image_rows}

    async def query_image_details_by_identifier(
        self,
        identifier: ImageIdentifier,
        status_filter: Optional[list[ImageStatus]] = None,
    ) -> ImageDataWithDetails:
        try:
            async with self._db.begin_readonly_session() as session:
                image_row = await ImageRow.resolve(
                    session,
                    [
                        identifier,
                        ImageAlias(identifier.canonical),
                    ],
                    filter_by_statuses=status_filter,
                )
        except UnknownImageReference:
            raise ImageNotFound
        data = image_row.to_detailed_dataclass()
        return data

    async def query_image_details_by_id(
        self,
        image_id: UUID,
        load_aliases: bool = False,
        status_filter: Optional[list[ImageStatus]] = None,
    ) -> ImageDataWithDetails:
        async with self._db.begin_session() as session:
            try:
                row: ImageRow = await self._get_image_by_id(
                    session, image_id, load_aliases, status_filter
                )
            except UnknownImageReference:
                raise ImageNotFound()
            data = row.to_detailed_dataclass()
        return data

    async def query_all_images(
        self, status_filter: Optional[list[ImageStatus]] = None
    ) -> Mapping[ImageID, ImageDataWithDetails]:
        async with self._db.begin_readonly_session() as session:
            rows = await ImageRow.list(session, load_aliases=True, filter_by_statuses=status_filter)
            return {ImageID(row.id): row.to_detailed_dataclass() for row in rows}

    async def mark_user_image_deleted(
        self,
        identifiers: list[ImageAlias | ImageRef | ImageIdentifier],
        user_id: UUID,
    ) -> ImageData:
        """
        Marks an image record as deleted for a specific user in the database.
        Raises ForgetImageActionGenericForbiddenError if the user does not own the image.
        """
        async with self._db.begin_session() as session:
            row = await self._resolve_image(session, identifiers)
            if not row.is_owned_by(user_id):
                raise ForgetImageForbiddenError()
            await row.mark_as_deleted(session)
            data = row.to_dataclass()
        return data

    async def mark_image_deleted_by_id(
        self,
        image_id: UUID,
        user_id: UUID,
    ) -> ImageData:
        """
        Marks an image record as deleted by its ID in the database.
        Validates ownership by user_id before deletion.
        Raises ForgetImageActionGenericForbiddenError if the user does not own the image.
        """
        async with self._db.begin_session() as session:
            image_row = await self._validate_image_ownership(session, image_id, user_id)
            await image_row.mark_as_deleted(session)
            data = image_row.to_dataclass()
        return data

    async def validate_and_fetch_image_ownership(
        self, image_id: UUID, user_id: UUID, load_aliases: bool = False
    ) -> ImageData:
        """
        Validates ownership and fetches an image from database by ID in a single operation.
        Raises ForgetImageActionGenericForbiddenError if image doesn't exist or user doesn't own it.
        """
        async with self._db.begin_session() as session:
            image_row = await self._validate_image_ownership(
                session, image_id, user_id, load_aliases
            )
            data = image_row.to_dataclass()
        return data

    async def insert_image_alias(
        self, alias: str, image_canonical: str, architecture: str
    ) -> tuple[UUID, ImageAliasData]:
        try:
            async with self._db.begin_session() as session:
                image_row = await ImageRow.resolve(
                    session, [ImageIdentifier(image_canonical, architecture)]
                )
                image_alias = ImageAliasRow(alias=alias, image_id=image_row.id)
                image_row.aliases.append(image_alias)
                row_id = image_row.id
                alias_data = ImageAliasData(id=image_alias.id, alias=image_alias.alias)
            return row_id, alias_data
        except ValueError:
            raise AliasImageActionValueError
        except DBAPIError as e:
            raise AliasImageActionDBError(str(e))

    async def query_image_alias(self, alias: str) -> ImageAliasData:
        async with self._db.begin_session() as session:
            row = await self._get_image_alias_by_name(session, alias)
            data = ImageAliasData(id=row.id, alias=row.alias)
        return data

    async def remove_image_alias(self, alias: str) -> tuple[UUID, ImageAliasData]:
        async with self._db.begin_session() as session:
            existing_alias = await self._get_image_alias_by_name(session, alias)
            image_id = existing_alias.image_id
            alias_data = ImageAliasData(id=existing_alias.id, alias=existing_alias.alias)
            await session.delete(existing_alias)
        return image_id, alias_data

    async def scan_and_upsert_image(
        self, image_canonical: str, architecture: str
    ) -> RescanImagesResult:
        """
        Scans a single image and upserts it into the database.
        Returns RescanImagesResult with the scanned image data.
        """

        async with self._db.begin_session() as session:
            # Resolve the image first
            image_row = await self._resolve_image(
                session, [ImageIdentifier(image_canonical, architecture)]
            )

            # Get the registry info
            registry_parts = []
            if image_row.registry:
                registry_parts.append(image_row.registry)
            if image_row.project:
                registry_parts.append(image_row.project)
            registry_key = "/".join(registry_parts) if registry_parts else ""

            # Get the registry row
            registry_row = await session.get(ContainerRegistryRow, image_row.registry_id)
            if not registry_row:
                raise RegistryNotFoundForImage(f"Registry not found for image {image_canonical}")

            # Call the original scan function
            result = await scan_single_image(self._db, registry_key, registry_row, image_canonical)

        return result

    async def remove_tag_from_registry(self, image_id: UUID) -> ImageData:
        async with self._db.begin_readonly_session() as session:
            image_row = await self._get_image_by_id(session, image_id, load_aliases=True)
            if not image_row:
                raise ImageNotFound()
            await image_row.untag_image_from_registry(self._db, session)
            data = image_row.to_dataclass()
        return data

    async def modify_image_properties(
        self, target: str, architecture: str, properties_to_update: dict
    ) -> ImageData:
        try:
            async with self._db.begin_session() as session:
                image_row = await ImageRow.resolve(
                    session,
                    [
                        ImageIdentifier(target, architecture),
                        ImageAlias(target),
                    ],
                )
                for key, value in properties_to_update.items():
                    setattr(image_row, key, value)
                data = image_row.to_dataclass()
            return data
        except (ValueError, DBAPIError):
            raise ModifyImageActionValueError

    async def clear_image_resource_limits(
        self, image_canonical: str, architecture: str
    ) -> ImageData:
        async with self._db.begin_session() as session:
            image_row = await ImageRow.resolve(
                session, [ImageIdentifier(image_canonical, architecture)]
            )
            image_row._resources = {}
            data = image_row.to_dataclass()
        return data

    async def remove_tag_from_registry_with_validation(
        self, image_id: UUID, user_id: UUID
    ) -> ImageData:
        """
        Validates ownership and removes an image registry tag in a single database operation.
        Raises ForgetImageActionGenericForbiddenError if user doesn't own the image.
        """
        async with self._db.begin_readonly_session() as session:
            image_row = await self._validate_image_ownership(
                session, image_id, user_id, load_aliases=True
            )
            await image_row.untag_image_from_registry(self._db, session)
            data = image_row.to_dataclass()
        return data

    async def remove_image_and_aliases_with_validation(
        self, image_id: UUID, user_id: UUID
    ) -> ImageData:
        """
        Removes an image record and all its aliases from database after validating ownership.
        Raises ForgetImageActionGenericForbiddenError if user doesn't own the image.
        """

        try:
            async with self._db.begin_session() as session:
                image_row = await self._validate_image_ownership(
                    session, image_id, user_id, load_aliases=True
                )
                data = image_row.to_dataclass()
                for alias in image_row.aliases:
                    await session.delete(alias)
                await session.delete(image_row)
            return data
        except DBAPIError as e:
            raise PurgeImageActionByIdObjectDBError(str(e))



"""
This module defines models that are used only when migrating existing data to the new RBAC system.
Any models defined here SHOULD NOT be used in the main RBAC system.

All models are defined in functional style.
The reason for this is to prevent some test code fixtures from incorrectly inferring SQLAlchemy model types.
"""

import sqlalchemy as sa
from sqlalchemy.orm import registry

from ai.backend.manager.models.base import GUID, IDColumn, metadata

mapper_registry = registry(metadata=metadata)


def get_user_roles_table() -> sa.Table:
    user_roles_table = sa.Table(
        "user_roles",
        mapper_registry.metadata,
        IDColumn(),
        sa.Column("user_id", GUID, nullable=False),
        sa.Column("role_id", GUID, nullable=False),
        extend_existing=True,
    )
    return user_roles_table


def get_roles_table() -> sa.Table:
    roles_table = sa.Table(
        "roles",
        mapper_registry.metadata,
        IDColumn(),
        sa.Column("name", sa.String(64), nullable=False),
        sa.Column(
            "source",
            sa.VARCHAR(16),
            nullable=False,
            default="system",
        ),
        sa.Column("description", sa.Text, nullable=True),
        extend_existing=True,
    )
    return roles_table


def get_permission_groups_table() -> sa.Table:
    permission_groups_table = sa.Table(
        "permission_groups",
        mapper_registry.metadata,
        IDColumn(),
        sa.Column("role_id", GUID, nullable=False),
        sa.Column("scope_type", sa.VARCHAR(length=32), nullable=False),
        sa.Column("scope_id", sa.String(length=64), nullable=False),
        extend_existing=True,
    )
    return permission_groups_table


def get_permissions_table() -> sa.Table:
    permissions_table = sa.Table(
        "permissions",
        mapper_registry.metadata,
        IDColumn(),
        sa.Column("permission_group_id", GUID, nullable=False),
        sa.Column("entity_type", sa.String(32), nullable=False),
        sa.Column("operation", sa.String(32), nullable=False),
        extend_existing=True,
    )
    return permissions_table


def get_object_permissions_table() -> sa.Table:
    object_permissions_table = sa.Table(
        "object_permissions",
        mapper_registry.metadata,
        IDColumn(),
        sa.Column("role_id", GUID, nullable=False),
        sa.Column("entity_type", sa.String(32), nullable=False),
        sa.Column("entity_id", sa.String(64), nullable=False),
        sa.Column("operation", sa.String(32), nullable=False),
        extend_existing=True,
    )
    return object_permissions_table


def get_association_scopes_entities_table() -> sa.Table:
    association_scopes_entities_table = sa.Table(
        "association_scopes_entities",
        mapper_registry.metadata,
        IDColumn(),
        sa.Column("scope_type", sa.String(32), nullable=False),
        sa.Column("scope_id", sa.String(64), nullable=False),
        sa.Column("entity_type", sa.String(32), nullable=False),
        sa.Column("entity_id", sa.String(64), nullable=False),
        extend_existing=True,
    )
    return association_scopes_entities_table



from __future__ import annotations

import uuid
from typing import Self

import strawberry
from strawberry import ID, Info
from strawberry.relay import Connection, Edge, Node, NodeID

from ai.backend.manager.api.gql.utils import dedent_strip
from ai.backend.manager.data.storage_namespace.creator import StorageNamespaceCreator
from ai.backend.manager.data.storage_namespace.types import StorageNamespaceData
from ai.backend.manager.services.storage_namespace.actions.register import (
    RegisterNamespaceAction,
)
from ai.backend.manager.services.storage_namespace.actions.unregister import (
    UnregisterNamespaceAction,
)

from .types import StrawberryGQLContext


@strawberry.type(
    description=dedent_strip("""
    Added in 25.15.0.

    Storage namespace provides logical separation of data within a single storage system
    to organize and isolate domain-specific concerns.

    Implementation varies by storage type:
    - Object Storage (S3, MinIO): Uses bucket-based namespace separation
    - File System (VFS): Uses directory path prefix for namespace distinction
    """)
)
class StorageNamespace(Node):
    id: NodeID[str]
    storage_id: ID
    namespace: str

    @classmethod
    def from_dataclass(cls, data: StorageNamespaceData) -> Self:
        return cls(
            id=ID(str(data.id)),
            storage_id=ID(str(data.storage_id)),
            namespace=data.namespace,
        )


StorageNamespaceEdge = Edge[StorageNamespace]


@strawberry.type(
    description=dedent_strip("""
    Added in 25.15.0.

    Storage namespace connection for pagination.
    """)
)
class StorageNamespaceConnection(Connection[StorageNamespace]):
    @strawberry.field
    def count(self) -> int:
        return len(self.edges)


@strawberry.input(
    description=dedent_strip("""
    Added in 25.15.0.

    Input type for registering a storage namespace.
    """)
)
class RegisterStorageNamespaceInput:
    storage_id: uuid.UUID
    namespace: str

    def to_creator(self) -> StorageNamespaceCreator:
        return StorageNamespaceCreator(
            storage_id=self.storage_id,
            bucket=self.namespace,
        )


@strawberry.input(
    description=dedent_strip("""
    Added in 25.15.0.

    Input type for unregistering a storage namespace.
    """)
)
class UnregisterStorageNamespaceInput:
    storage_id: uuid.UUID
    namespace: str


@strawberry.type(
    description=dedent_strip("""
    Added in 25.15.0.

    Payload returned after storage namespace registration.
    """)
)
class RegisterStorageNamespacePayload:
    id: uuid.UUID


@strawberry.type(
    description=dedent_strip("""
    Added in 25.15.0.

    Payload returned after storage namespace unregistration.
    """)
)
class UnregisterStorageNamespacePayload:
    id: uuid.UUID


@strawberry.mutation(
    description=dedent_strip("""
    Added in 25.15.0.

    Registers a new namespace within a storage.
    """)
)
async def register_storage_namespace(
    input: RegisterStorageNamespaceInput, info: Info[StrawberryGQLContext]
) -> RegisterStorageNamespacePayload:
    processors = info.context.processors

    action_result = await processors.storage_namespace.register.wait_for_complete(
        RegisterNamespaceAction(
            creator=input.to_creator(),
        )
    )

    return RegisterStorageNamespacePayload(id=action_result.storage_id)


@strawberry.mutation(
    description=dedent_strip("""
    Added in 25.15.0.

    Unregisters an existing namespace from a storage.
    """)
)
async def unregister_storage_namespace(
    input: UnregisterStorageNamespaceInput, info: Info[StrawberryGQLContext]
) -> UnregisterStorageNamespacePayload:
    processors = info.context.processors

    action_result = await processors.storage_namespace.unregister.wait_for_complete(
        UnregisterNamespaceAction(
            storage_id=input.storage_id,
            namespace=input.namespace,
        )
    )

    return UnregisterStorageNamespacePayload(id=action_result.storage_id)



from collections.abc import Mapping
from dataclasses import dataclass
from typing import Optional, override

from ai.backend.common.types import AgentId, ImageID
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.services.image.actions.base import ImageAction


@dataclass
class GetImageInstalledAgentsAction(ImageAction):
    image_ids: list[ImageID]

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "get_installed_agents"


@dataclass
class GetImageInstalledAgentsActionResult(BaseActionResult):
    data: Mapping[ImageID, set[AgentId]]

    @override
    def entity_id(self) -> Optional[str]:
        return None



import uuid
from http import HTTPStatus
from typing import Optional, Protocol

from ai.backend.common.api_handlers import (
    APIResponse,
    BodyParam,
    PathParam,
    QueryParam,
    api_handler,
)
from ai.backend.common.dto.manager.context import KeypairCtx, UserIdentityCtx
from ai.backend.common.dto.manager.path import VFolderIDPath
from ai.backend.common.dto.manager.query import ListGroupQuery
from ai.backend.common.dto.manager.request import (
    RenameVFolderReq,
    VFolderCreateReq,
)
from ai.backend.common.dto.manager.response import VFolderCreateResponse, VFolderListResponse
from ai.backend.manager.data.vfolder.dto import (
    Keypair,
    UserIdentity,
    VFolderItem,
    VFolderItemToCreate,
)


class VFolderServiceProtocol(Protocol):
    async def create_vfolder_in_personal(
        self,
        user_identity: UserIdentity,
        keypair: Keypair,
        vfolder_info: VFolderItemToCreate,
    ) -> VFolderItem: ...

    async def create_vfolder_in_group(
        self,
        user_identity: UserIdentity,
        keypair: Keypair,
        vfolder_info: VFolderItemToCreate,
    ) -> VFolderItem: ...

    async def get_vfolders(
        self, user_identity: UserIdentity, group_id: Optional[uuid.UUID]
    ) -> list[VFolderItem]: ...

    async def rename_vfolder(
        self, user_identity: UserIdentity, keypair: Keypair, vfolder_id: uuid.UUID, new_name: str
    ) -> None: ...

    async def delete_vfolder(
        self,
        vfolder_id: uuid.UUID,
        user_identity: UserIdentity,
        keypair: Keypair,
    ) -> None: ...


class VFolderHandler:
    _vfolder_service: VFolderServiceProtocol

    def __init__(self, vfolder_service: VFolderServiceProtocol):
        self._vfolder_service = vfolder_service

    @api_handler
    async def create_vfolder(
        self,
        keypair_ctx: KeypairCtx,
        user_identity_ctx: UserIdentityCtx,
        body: BodyParam[VFolderCreateReq],
    ) -> APIResponse:
        vfolder_info = await VFolderItemToCreate.from_request(body.parsed)
        user_identity = UserIdentity.from_ctx(user_identity_ctx)
        keypair = Keypair.from_ctx(keypair_ctx)
        created_vfolder: VFolderItem
        if vfolder_info.group_id:
            created_vfolder = await self._vfolder_service.create_vfolder_in_group(
                user_identity=user_identity,
                keypair=keypair,
                vfolder_info=vfolder_info,
            )
        else:
            created_vfolder = await self._vfolder_service.create_vfolder_in_personal(
                user_identity=user_identity,
                keypair=keypair,
                vfolder_info=vfolder_info,
            )

        return APIResponse.build(
            status_code=HTTPStatus.OK,
            response_model=VFolderCreateResponse(item=created_vfolder.to_field()),
        )

    @api_handler
    async def list_vfolders(
        self, user_identity_ctx: UserIdentityCtx, query: QueryParam[ListGroupQuery]
    ) -> APIResponse:
        list_group_query = query.parsed
        user_identity = UserIdentity.from_ctx(user_identity_ctx)

        vfolder_list: list[VFolderItem] = await self._vfolder_service.get_vfolders(
            user_identity=user_identity, group_id=list_group_query.group_id
        )

        return APIResponse.build(
            status_code=HTTPStatus.OK,
            response_model=VFolderListResponse(
                items=[vfolder.to_field() for vfolder in vfolder_list]
            ),
        )

    @api_handler
    async def rename_vfolder(
        self,
        keypair_ctx: KeypairCtx,
        user_identity_ctx: UserIdentityCtx,
        path: PathParam[VFolderIDPath],
        body: BodyParam[RenameVFolderReq],
    ) -> APIResponse:
        vfolder_id: uuid.UUID = path.parsed.vfolder_id
        new_name: str = body.parsed.new_name
        user_identity = UserIdentity.from_ctx(user_identity_ctx)
        keypair = Keypair.from_ctx(keypair_ctx)

        await self._vfolder_service.rename_vfolder(
            user_identity=user_identity,
            keypair=keypair,
            vfolder_id=vfolder_id,
            new_name=new_name,
        )
        return APIResponse.no_content(status_code=HTTPStatus.CREATED)

    @api_handler
    async def delete_vfolder(
        self,
        keypair_ctx: KeypairCtx,
        user_identity_ctx: UserIdentityCtx,
        path: PathParam[VFolderIDPath],
    ) -> APIResponse:
        parsed_path = path.parsed
        user_identity = UserIdentity.from_ctx(user_identity_ctx)
        keypair = Keypair.from_ctx(keypair_ctx)

        await self._vfolder_service.delete_vfolder(
            user_identity=user_identity,
            keypair=keypair,
            vfolder_id=parsed_path.vfolder_id,
        )
        return APIResponse.no_content(status_code=HTTPStatus.NO_CONTENT)



from dataclasses import dataclass
from typing import Self

from ai.backend.manager.repositories.artifact_registry.repository import ArtifactRegistryRepository
from ai.backend.manager.repositories.types import RepositoryArgs


@dataclass
class ArtifactRegistryRepositories:
    repository: ArtifactRegistryRepository

    @classmethod
    def create(cls, args: RepositoryArgs) -> Self:
        repository = ArtifactRegistryRepository(args.db)

        return cls(
            repository=repository,
        )



from __future__ import annotations

import asyncio
import enum
import functools
import logging
import sys
import uuid
from collections.abc import (
    Awaitable,
    Callable,
    Iterable,
    Mapping,
    MutableMapping,
    Sequence,
)
from decimal import Decimal
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Concatenate,
    Final,
    Generic,
    NamedTuple,
    Optional,
    Protocol,
    Self,
    TypeVar,
    cast,
    overload,
)

import graphene
import sqlalchemy as sa
import trafaret as t
import yarl
from aiodataloader import DataLoader
from aiotools import apartial
from dateutil.parser import isoparse
from graphene.types import Scalar
from graphene.types.scalars import MAX_INT, MIN_INT
from graphene_federation import shareable
from graphql import Undefined
from graphql.language.ast import IntValueNode
from sqlalchemy.dialects.postgresql import ARRAY, CIDR, ENUM, JSONB, UUID
from sqlalchemy.engine.result import Result
from sqlalchemy.engine.row import Row
from sqlalchemy.ext.asyncio import AsyncConnection as SAConnection
from sqlalchemy.ext.asyncio import AsyncEngine as SAEngine
from sqlalchemy.ext.asyncio import AsyncSession as SASession
from sqlalchemy.orm import DeclarativeMeta, registry
from sqlalchemy.types import CHAR, SchemaType, TypeDecorator, Unicode, UnicodeText

from ai.backend.common import validators as tx
from ai.backend.common.auth import PublicKey
from ai.backend.common.exception import InvalidIpAddressValue
from ai.backend.common.types import (
    AbstractPermission,
    EndpointId,
    JSONSerializableMixin,
    KernelId,
    QuotaScopeID,
    ReadableCIDR,
    ResourceSlot,
    SessionId,
    VFolderHostPermission,
    VFolderHostPermissionMap,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.auth.hash import PasswordHashAlgorithm
from ai.backend.manager.models.hasher.types import PasswordInfo

from ..errors.api import InvalidAPIParameters
from ..errors.common import GenericForbidden
from ..errors.resource import DataTransformationFailed
from .gql_relay import (
    AsyncListConnectionField,
    AsyncNode,
    ConnectionPaginationOrder,
)
from .minilang.ordering import OrderDirection, OrderingItem, QueryOrderParser
from .minilang.queryfilter import QueryFilterParser, WhereClauseType
from .utils import execute_with_retry

if TYPE_CHECKING:
    from sqlalchemy.engine.interfaces import Dialect
    from sqlalchemy.orm.attributes import InstrumentedAttribute
    from sqlalchemy.sql.selectable import ScalarSelect

    from .gql import GraphQueryContext
    from .user import UserRole

SAFE_MIN_INT = -9007199254740991
SAFE_MAX_INT = 9007199254740991

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

# The common shared metadata instance
convention = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}
metadata = sa.MetaData(naming_convention=convention)
mapper_registry = registry(metadata=metadata)
Base: Any = mapper_registry.generate_base()  # TODO: remove Any after #422 is merged

pgsql_connect_opts = {
    "server_settings": {
        "jit": "off",
        # 'deadlock_timeout': '10000',  # FIXME: AWS RDS forbids settings this via connection arguments
        "lock_timeout": "60000",  # 60 secs
        "idle_in_transaction_session_timeout": "60000",  # 60 secs
    },
}

DEFAULT_PAGE_SIZE: Final[int] = 10


# helper functions
def zero_if_none(val):
    return 0 if val is None else val


class FixtureOpModes(enum.StrEnum):
    INSERT = "insert"
    UPDATE = "update"


T_Enum = TypeVar("T_Enum", bound=enum.Enum, covariant=True)
T_StrEnum = TypeVar("T_StrEnum", bound=enum.Enum, covariant=True)


class EnumType(TypeDecorator, SchemaType, Generic[T_Enum]):
    """
    A stripped-down version of Spoqa's sqlalchemy-enum34.
    It also handles postgres-specific enum type creation.

    The actual postgres enum choices are taken from the Python enum names.
    """

    impl = ENUM
    cache_ok = True

    def __init__(self, enum_cls: type[T_Enum], **opts) -> None:
        if "name" not in opts:
            opts["name"] = enum_cls.__name__.lower()
        self._opts = opts
        enums = (m.name for m in enum_cls)
        super().__init__(*enums, **opts)
        self._enum_cls = enum_cls

    def process_bind_param(
        self,
        value: Optional[T_Enum],
        dialect: Dialect,
    ) -> Optional[str]:
        return value.name if value else None

    def process_result_value(
        self,
        value: str,
        dialect: Dialect,
    ) -> Optional[T_Enum]:
        return self._enum_cls[value] if value else None

    def copy(self, **kw) -> type[Self]:
        return EnumType(self._enum_cls, **self._opts)

    @property
    def python_type(self):
        return self._enum_class


class EnumValueType(TypeDecorator, SchemaType, Generic[T_Enum]):
    """
    A stripped-down version of Spoqa's sqlalchemy-enum34.
    It also handles postgres-specific enum type creation.

    The actual postgres enum choices are taken from the Python enum values.
    """

    impl = ENUM
    cache_ok = True

    def __init__(self, enum_cls: type[T_Enum], **opts) -> None:
        if "name" not in opts:
            opts["name"] = enum_cls.__name__.lower()
        self._opts = opts
        enums = (m.value for m in enum_cls)
        super().__init__(*enums, **opts)
        self._enum_cls = enum_cls

    def process_bind_param(
        self,
        value: Optional[T_Enum],
        dialect: Dialect,
    ) -> Optional[str]:
        return value.value if value else None

    def process_result_value(
        self,
        value: str,
        dialect: Dialect,
    ) -> Optional[T_Enum]:
        return self._enum_cls(value) if value else None

    def copy(self, **kw) -> type[Self]:
        return EnumValueType(self._enum_cls, **self._opts)

    @property
    def python_type(self) -> T_Enum:
        return self._enum_class


class StrEnumType(TypeDecorator, Generic[T_StrEnum]):
    """
    Maps Postgres VARCHAR(64) column with a Python enum.StrEnum type.
    """

    impl = sa.VARCHAR
    cache_ok = True

    def __init__(
        self, enum_cls: type[T_StrEnum], use_name: bool = False, length: int = 64, **opts
    ) -> None:
        self._opts = opts
        super().__init__(length=length, **opts)
        self._use_name = use_name
        self._enum_cls = enum_cls

    def process_bind_param(
        self,
        value: Optional[T_StrEnum],
        dialect: Dialect,
    ) -> Optional[str]:
        if value is None:
            return None
        if self._use_name:
            return value.name
        else:
            return value.value

    def process_result_value(
        self,
        value: Optional[str],
        dialect: Dialect,
    ) -> Optional[T_StrEnum]:
        if value is None:
            return None
        if self._use_name:
            return self._enum_cls[value]
        else:
            return self._enum_cls(value)

    def copy(self, **kw) -> type[Self]:
        return StrEnumType(self._enum_cls, self._use_name, **self._opts)

    @property
    def python_type(self) -> type[T_StrEnum]:
        return self._enum_cls


class CurvePublicKeyColumn(TypeDecorator):
    """
    A column type wrapper for string-based Z85-encoded CURVE public key.

    In the database, it resides as a string but it's safe to just convert them to PublicKey (bytes)
    because zmq uses Z85 encoding to use printable characters only in the key while the pyzmq API
    treats them as bytes.

    The pure binary representation of a public key is 32 bytes and its Z85-encoded form used in the
    pyzmq APIs is 40 ASCII characters.
    """

    impl = sa.String
    cache_ok = True

    def load_dialect_impl(self, dialect):
        return dialect.type_descriptor(sa.String(40))

    def process_bind_param(
        self,
        value: Optional[PublicKey],
        dialect: Dialect,
    ) -> Optional[str]:
        return value.decode("ascii") if value else None

    def process_result_value(
        self,
        value: str | None,
        dialect: Dialect,
    ) -> Optional[PublicKey]:
        if value is None:
            return None
        return PublicKey(value.encode("ascii"))


class QuotaScopeIDType(TypeDecorator):
    """
    A column type wrapper for string-based quota scope ID.
    """

    impl = sa.String
    cache_ok = True

    def load_dialect_impl(self, dialect):
        return dialect.type_descriptor(sa.String(64))

    def process_bind_param(
        self,
        value: Optional[QuotaScopeID],
        dialect: Dialect,
    ) -> Optional[str]:
        return str(value) if value else None

    def process_result_value(
        self,
        value: Optional[str],
        dialect: Dialect,
    ) -> Optional[QuotaScopeID]:
        return QuotaScopeID.parse(value) if value else None


class ResourceSlotColumn(TypeDecorator):
    """
    A column type wrapper for ResourceSlot from JSONB.
    """

    impl = JSONB
    cache_ok = True

    def process_bind_param(
        self,
        value: Optional[ResourceSlot],
        dialect: Dialect,
    ) -> Optional[Mapping[str, str]]:
        if value is None:
            return None
        if isinstance(value, ResourceSlot):
            return value.to_json()
        return value

    def process_result_value(
        self,
        value: Optional[dict[str, str]],
        dialect: Dialect,
    ) -> Optional[ResourceSlot]:
        if value is None:
            return None
        try:
            return ResourceSlot.from_json(value)
        except ArithmeticError:
            # for legacy-compat scenario
            return ResourceSlot.from_user_input(value, None)


class StructuredJSONColumn(TypeDecorator):
    """
    A column type to convert JSON values back and forth using a Trafaret.
    """

    impl = JSONB
    cache_ok = True

    def __init__(self, schema: t.Trafaret) -> None:
        super().__init__()
        self._schema = schema

    def load_dialect_impl(self, dialect: Dialect):
        if dialect.name == "sqlite":
            return dialect.type_descriptor(sa.JSON)
        else:
            return super().load_dialect_impl(dialect)

    def process_bind_param(
        self,
        value: Optional[Any],
        dialect: Dialect,
    ) -> Optional[Any]:
        if value is None:
            return self._schema.check({})
        try:
            self._schema.check(value)
        except t.DataError as e:
            raise ValueError(
                "The given value does not conform with the structured json column format.",
                e.as_dict(),
            )
        return value

    def process_result_value(
        self,
        value: Optional[Any],
        dialect: Dialect,
    ) -> Optional[Any]:
        if value is None:
            return self._schema.check({})
        return self._schema.check(value)

    def copy(self, **kw) -> type[Self]:
        return StructuredJSONColumn(self._schema)


class StructuredJSONObjectColumn(TypeDecorator):
    """
    A column type to convert JSON values back and forth using JSONSerializableMixin.
    """

    impl = JSONB
    cache_ok = True

    def __init__(self, schema: type[JSONSerializableMixin]) -> None:
        super().__init__()
        self._schema = schema

    def process_bind_param(self, value, dialect):
        return self._schema.to_json(value)

    def process_result_value(self, value, dialect):
        return self._schema.from_json(value)

    def copy(self, **kw) -> type[Self]:
        return StructuredJSONObjectColumn(self._schema)


class StructuredJSONObjectListColumn(TypeDecorator):
    """
    A column type to convert JSON values back and forth using JSONSerializableMixin,
    but store and load a list of the objects.
    """

    impl = JSONB
    cache_ok = True

    def __init__(self, schema: type[JSONSerializableMixin]) -> None:
        super().__init__()
        self._schema = schema

    def coerce_compared_value(self, op, value):
        return JSONB()

    def process_bind_param(self, value, dialect):
        return [self._schema.to_json(item) for item in value]

    def process_result_value(self, value, dialect):
        if value is None:
            return []
        return [self._schema.from_json(item) for item in value]

    def copy(self, **kw) -> type[Self]:
        return StructuredJSONObjectListColumn(self._schema)


class URLColumn(TypeDecorator):
    """
    A column type for URL strings
    """

    impl = UnicodeText
    cache_ok = True

    def process_bind_param(self, value: Optional[yarl.URL], dialect: Dialect) -> Optional[str]:
        return str(value)

    def process_result_value(self, value: Optional[str], dialect: Dialect) -> Optional[yarl.URL]:
        if value is None:
            return None
        if value is not None:
            return yarl.URL(value)


class IPColumn(TypeDecorator):
    """
    A column type to convert IP string values back and forth to CIDR.
    """

    impl = CIDR
    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        try:
            cidr = ReadableCIDR(value).address
        except InvalidIpAddressValue:
            raise InvalidAPIParameters(f"{value} is invalid IP address value")
        return cidr

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        return ReadableCIDR(value)


class PermissionListColumn(TypeDecorator):
    """
    A column type to convert Permission values back and forth.
    """

    impl = ARRAY
    cache_ok = True

    def __init__(self, perm_type: type[AbstractPermission]) -> None:
        super().__init__(sa.String)
        self._perm_type = perm_type

    @overload
    def process_bind_param(
        self, value: Sequence[AbstractPermission], dialect: Dialect
    ) -> list[str]: ...

    @overload
    def process_bind_param(self, value: Sequence[str], dialect: Dialect) -> list[str]: ...

    @overload
    def process_bind_param(self, value: None, dialect: Dialect) -> list[str]: ...

    def process_bind_param(
        self,
        value: Sequence[AbstractPermission] | Sequence[str] | None,
        dialect: Dialect,
    ) -> list[str]:
        if value is None:
            return []
        try:
            return [self._perm_type(perm).value for perm in value]
        except ValueError:
            raise InvalidAPIParameters(f"Invalid value for binding to {self._perm_type}")

    def process_result_value(
        self,
        value: Sequence[str] | None,
        dialect: Dialect,
    ) -> set[AbstractPermission]:
        if value is None:
            return set()
        return set(self._perm_type(perm) for perm in value)


class VFolderHostPermissionColumn(TypeDecorator):
    """
    A column type to convert vfolder host permission back and forth.
    """

    impl = JSONB
    cache_ok = True
    perm_col = PermissionListColumn(VFolderHostPermission)

    def process_bind_param(
        self,
        value: Mapping[str, Any] | None,
        dialect: Dialect,
    ) -> Mapping[str, Any]:
        if value is None:
            return {}
        return {
            host: self.perm_col.process_bind_param(perms, None) for host, perms in value.items()
        }

    def process_result_value(
        self,
        value: Mapping[str, Any] | None,
        dialect: Dialect,
    ) -> VFolderHostPermissionMap:
        if value is None:
            return VFolderHostPermissionMap()
        return VFolderHostPermissionMap({
            host: self.perm_col.process_result_value(perms, None) for host, perms in value.items()
        })


class CurrencyTypes(enum.Enum):
    KRW = "KRW"
    USD = "USD"


TUUIDSubType = TypeVar("TUUIDSubType", bound=uuid.UUID)


class GUID(TypeDecorator, Generic[TUUIDSubType]):
    """
    Platform-independent GUID type.
    Uses PostgreSQL's UUID type, otherwise uses CHAR(16) storing as raw bytes.
    """

    impl = CHAR
    uuid_subtype_func: ClassVar[Callable[[Any], uuid.UUID]] = lambda v: v
    cache_ok = True

    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql":
            return dialect.type_descriptor(UUID())
        else:
            return dialect.type_descriptor(CHAR(16))

    def process_bind_param(self, value: TUUIDSubType | uuid.UUID, dialect):
        # NOTE: EndpointId, SessionId, KernelId are *not* actual types defined as classes,
        #       but a "virtual" type that is an identity function at runtime.
        #       The type checker treats them as distinct derivatives of uuid.UUID.
        #       Therefore, we just do isinstance on uuid.UUID only below.
        if value is None:
            return value
        elif dialect.name == "postgresql":
            if isinstance(value, uuid.UUID):
                return str(value)
            else:
                return str(uuid.UUID(value))
        else:
            if isinstance(value, uuid.UUID):
                return value.bytes
            else:
                return uuid.UUID(value).bytes

    def process_result_value(self, value: Any, dialect) -> Optional[TUUIDSubType]:
        if value is None:
            return value
        else:
            cls = type(self)
            if isinstance(value, bytes):
                return cast(TUUIDSubType, cls.uuid_subtype_func(uuid.UUID(bytes=value)))
            else:
                return cast(TUUIDSubType, cls.uuid_subtype_func(uuid.UUID(value)))


class SlugType(TypeDecorator):
    """
    A type wrapper for slug type string
    """

    impl = Unicode
    cache_ok = True

    def __init__(
        self,
        *,
        length: int | None = None,
        allow_dot: bool = False,
        allow_space: bool = False,
        allow_unicode: bool = False,
    ) -> None:
        super().__init__(length=length)
        self._tx_slug = tx.Slug(
            max_length=length,
            allow_dot=allow_dot,
            allow_space=allow_space,
            allow_unicode=allow_unicode,
        )

    def coerce_compared_value(self, op, value):
        return Unicode()

    def process_bind_param(self, value: str, dialect) -> str:
        try:
            self._tx_slug.check(value)
        except t.DataError as e:
            raise ValueError(e.error, value)
        return value


class EndpointIDColumnType(GUID[EndpointId]):
    uuid_subtype_func = lambda v: EndpointId(v)
    cache_ok = True


class SessionIDColumnType(GUID[SessionId]):
    uuid_subtype_func = lambda v: SessionId(v)
    cache_ok = True


class KernelIDColumnType(GUID[KernelId]):
    uuid_subtype_func = lambda v: KernelId(v)
    cache_ok = True


def IDColumn(name="id"):
    return sa.Column(name, GUID, primary_key=True, server_default=sa.text("uuid_generate_v4()"))


def EndpointIDColumn(name="id"):
    return sa.Column(
        name, EndpointIDColumnType, primary_key=True, server_default=sa.text("uuid_generate_v4()")
    )


def SessionIDColumn(name="id"):
    return sa.Column(
        name, SessionIDColumnType, primary_key=True, server_default=sa.text("uuid_generate_v4()")
    )


def KernelIDColumn(name="id"):
    return sa.Column(
        name, KernelIDColumnType, primary_key=True, server_default=sa.text("uuid_generate_v4()")
    )


def ForeignKeyIDColumn(name, fk_field, nullable=True):
    return sa.Column(name, GUID, sa.ForeignKey(fk_field), nullable=nullable)


TContext = TypeVar("TContext")
TLoaderKey = TypeVar("TLoaderKey")
TLoaderResult = TypeVar("TLoaderResult")


class DataLoaderManager(Generic[TContext, TLoaderKey, TLoaderResult]):
    """
    For every different combination of filtering conditions, we need to make a
    new DataLoader instance because it "batches" the database queries.
    This manager get-or-creates dataloaders with fixed conditions (represetned
    as arguments) like a cache.

    NOTE: Just like DataLoaders, it is recommended to instantiate this manager
    for every incoming API request.
    """

    cache: dict[int, DataLoader[TLoaderKey, TLoaderResult]]

    def __init__(self) -> None:
        self.cache = {}
        self.mod = sys.modules["ai.backend.manager.models"]
        self._gql_mod = sys.modules["ai.backend.manager.models.gql_models"]

    @staticmethod
    def _get_key(otname: str, args, kwargs) -> int:
        """
        Calculate the hash of the all arguments and keyword arguments.
        """
        key = (otname,) + args
        for item in kwargs.items():
            key += item
        return hash(key)

    # TODO: Remove this method and logic to parse `batch_load_` method name
    def load_attr(self, objtype_name: str) -> Any:
        try:
            return getattr(self.mod, objtype_name)
        except Exception:
            return getattr(self._gql_mod, objtype_name)

    def get_loader(
        self, context: GraphQueryContext, objtype_name: str, *args, **kwargs
    ) -> DataLoader:
        k = self._get_key(objtype_name, args, kwargs)
        loader = self.cache.get(k)
        if loader is None:
            objtype_name, has_variant, variant_name = objtype_name.partition(".")
            objtype = self.load_attr(objtype_name)
            if has_variant:
                batch_load_fn = getattr(objtype, "batch_load_" + variant_name)
            else:
                batch_load_fn = objtype.batch_load
            loader = DataLoader(
                apartial(batch_load_fn, context, *args, **kwargs),
                max_batch_size=128,
            )
            self.cache[k] = loader
        return loader

    @staticmethod
    def _get_func_key(
        func: Callable[
            Concatenate[TContext, Sequence[TLoaderKey], ...],
            Awaitable[Sequence[TLoaderResult]],
        ],
        **kwargs,
    ) -> int:
        func_and_kwargs = (func, *[(k, kwargs[k]) for k in sorted(kwargs.keys())])
        return hash(func_and_kwargs)

    def get_loader_by_func(
        self,
        context: TContext,
        batch_load_func: Callable[
            Concatenate[TContext, Sequence[TLoaderKey], ...],
            Awaitable[Sequence[TLoaderResult]],
        ],
        # Using kwargs-only to prevent argument position confusion
        # when DataLoader calls `batch_load_func(keys)` which is `partial(batch_load_func, **kwargs)(keys)`.
        **kwargs,
    ) -> DataLoader[TLoaderKey, TLoaderResult]:
        async def batch_load_wrapper(keys: Sequence[TLoaderKey]) -> list[TLoaderResult]:
            # aiodataloader always converts the result via list(),
            # so we can enforce type-casting here.
            return cast(list[TLoaderResult], await batch_load_func(context, keys, **kwargs))

        key = self._get_func_key(batch_load_func, **kwargs)
        loader = self.cache.get(key)
        if loader is None:
            loader = DataLoader(
                batch_load_wrapper,
                max_batch_size=128,
            )
            self.cache[key] = loader
        return loader


@shareable
class ResourceLimit(graphene.ObjectType):
    key = graphene.String()
    min = graphene.String()
    max = graphene.String(
        deprecation_reason="Deprecated since 25.14.0. The max slot limit validation has been removed as it was deemed obsolete."
    )


@shareable
class KVPair(graphene.ObjectType):
    key = graphene.String()
    value = graphene.String()


class ResourceLimitInput(graphene.InputObjectType):
    key = graphene.String()
    min = graphene.String()
    max = graphene.String(
        deprecation_reason="Deprecated since 25.14.0. The max slot limit validation has been removed as it was deemed obsolete."
    )


class KVPairInput(graphene.InputObjectType):
    key = graphene.String()
    value = graphene.String()


class BigInt(Scalar):
    """
    BigInt is an extension of the regular graphene.Int scalar type
    to support integers outside the range of a signed 32-bit integer.
    """

    @staticmethod
    def coerce_bigint(value):
        num = int(value)
        if not (SAFE_MIN_INT <= num <= SAFE_MAX_INT):
            raise ValueError("Cannot serialize integer out of the safe range.")
        if not (MIN_INT <= num <= MAX_INT):
            # treat as float
            return float(int(num))
        return num

    serialize = coerce_bigint
    parse_value = coerce_bigint

    @staticmethod
    def parse_literal(node):
        if isinstance(node, IntValueNode):
            num = int(node.value)
            if not (SAFE_MIN_INT <= num <= SAFE_MAX_INT):
                raise ValueError("Cannot parse integer out of the safe range.")
            if not (MIN_INT <= num <= MAX_INT):
                # treat as float
                return float(int(num))
            return num


class Item(graphene.Interface):
    id = graphene.ID()


class PaginatedList(graphene.Interface):
    items = graphene.List(Item, required=True)
    total_count = graphene.Int(required=True)


# ref: https://github.com/python/mypy/issues/1212
T_SQLBasedGQLObject = TypeVar("T_SQLBasedGQLObject", bound="_SQLBasedGQLObject")
T_Key = TypeVar("T_Key")


class _SQLBasedGQLObject(Protocol):
    @classmethod
    def from_row(
        cls: type[T_SQLBasedGQLObject],
        ctx: GraphQueryContext,
        row: Row | DeclarativeMeta,
    ) -> T_SQLBasedGQLObject: ...


async def batch_result(
    graph_ctx: GraphQueryContext,
    db_conn: SAConnection | SASession,
    query: sa.sql.Select,
    obj_type: type[T_SQLBasedGQLObject],
    key_list: Iterable[T_Key],
    key_getter: Callable[[Row], T_Key],
) -> Sequence[Optional[T_SQLBasedGQLObject]]:
    """
    A batched query adaptor for (key -> item) resolving patterns.
    """
    objs_per_key: dict[T_Key, Optional[T_SQLBasedGQLObject]]
    objs_per_key = dict()
    for key in key_list:
        objs_per_key[key] = None
    if isinstance(db_conn, SASession):
        stream_func = db_conn.stream_scalars
    else:
        stream_func = db_conn.stream
    async for row in await stream_func(query):
        objs_per_key[key_getter(row)] = obj_type.from_row(graph_ctx, row)
    return [*objs_per_key.values()]


async def batch_multiresult(
    graph_ctx: GraphQueryContext,
    db_conn: SAConnection | SASession,
    query: sa.sql.Select,
    obj_type: type[T_SQLBasedGQLObject],
    key_list: Iterable[T_Key],
    key_getter: Callable[[Row], T_Key],
) -> Sequence[Sequence[T_SQLBasedGQLObject]]:
    """
    A batched query adaptor for (key -> [item]) resolving patterns.
    """
    objs_per_key: dict[T_Key, list[T_SQLBasedGQLObject]]
    objs_per_key = dict()
    for key in key_list:
        objs_per_key[key] = list()
    if isinstance(db_conn, SASession):
        stream_func = db_conn.stream_scalars
    else:
        stream_func = db_conn.stream
    async for row in await stream_func(query):
        objs_per_key[key_getter(row)].append(
            obj_type.from_row(graph_ctx, row),
        )
    return [*objs_per_key.values()]


async def batch_result_in_session(
    graph_ctx: GraphQueryContext,
    db_sess: SASession,
    query: sa.sql.Select,
    obj_type: type[T_SQLBasedGQLObject],
    key_list: Iterable[T_Key],
    key_getter: Callable[[Row], T_Key],
) -> Sequence[Optional[T_SQLBasedGQLObject]]:
    """
    A batched query adaptor for (key -> item) resolving patterns.
    stream the result in async session.
    """
    objs_per_key: dict[T_Key, Optional[T_SQLBasedGQLObject]]
    objs_per_key = dict()
    for key in key_list:
        objs_per_key[key] = None
    async for row in await db_sess.stream(query):
        objs_per_key[key_getter(row)] = obj_type.from_row(graph_ctx, row)
    return [*objs_per_key.values()]


async def batch_result_in_scalar_stream(
    graph_ctx: GraphQueryContext,
    db_sess: SASession,
    query: sa.sql.Select,
    obj_type: type[T_SQLBasedGQLObject],
    key_list: Iterable[T_Key],
    key_getter: Callable[[Row], T_Key],
) -> Sequence[Optional[T_SQLBasedGQLObject]]:
    """
    A batched query adaptor for (key -> item) resolving patterns.
    stream the result scalar in async session.
    """
    objs_per_key: dict[T_Key, Optional[T_SQLBasedGQLObject]]
    objs_per_key = {}
    for key in key_list:
        objs_per_key[key] = None
    async for row in await db_sess.stream_scalars(query):
        objs_per_key[key_getter(row)] = obj_type.from_row(graph_ctx, row)
    return [*objs_per_key.values()]


async def batch_multiresult_in_session(
    graph_ctx: GraphQueryContext,
    db_sess: SASession,
    query: sa.sql.Select,
    obj_type: type[T_SQLBasedGQLObject],
    key_list: Iterable[T_Key],
    key_getter: Callable[[Row], T_Key],
) -> Sequence[Sequence[T_SQLBasedGQLObject]]:
    """
    A batched query adaptor for (key -> [item]) resolving patterns.
    stream the result in async session.
    """
    objs_per_key: dict[T_Key, list[T_SQLBasedGQLObject]]
    objs_per_key = dict()
    for key in key_list:
        objs_per_key[key] = list()
    async for row in await db_sess.stream(query):
        objs_per_key[key_getter(row)].append(
            obj_type.from_row(graph_ctx, row),
        )
    return [*objs_per_key.values()]


async def batch_multiresult_in_scalar_stream(
    graph_ctx: GraphQueryContext,
    db_sess: SASession,
    query: sa.sql.Select,
    obj_type: type[T_SQLBasedGQLObject],
    key_list: Iterable[T_Key],
    key_getter: Callable[[Row], T_Key],
) -> Sequence[Sequence[T_SQLBasedGQLObject]]:
    """
    A batched query adaptor for (key -> [item]) resolving patterns.
    stream the result in async session.
    """
    objs_per_key: dict[T_Key, list[T_SQLBasedGQLObject]]
    objs_per_key = dict()
    for key in key_list:
        objs_per_key[key] = list()
    async for row in await db_sess.stream_scalars(query):
        objs_per_key[key_getter(row)].append(
            obj_type.from_row(graph_ctx, row),
        )
    return [*objs_per_key.values()]


def privileged_query(required_role: UserRole):
    def wrap(func):
        @functools.wraps(func)
        async def wrapped(
            root: Any,
            info: graphene.ResolveInfo,
            *args,
            **kwargs,
        ) -> Any:
            from .user import UserRole

            ctx: GraphQueryContext = info.context
            if ctx.user["role"] != UserRole.SUPERADMIN:
                raise GenericForbidden("superadmin privilege required")
            return await func(root, info, *args, **kwargs)

        return wrapped

    return wrap


def scoped_query(
    *,
    autofill_user: bool = False,
    user_key: str = "access_key",
):
    """
    Prepends checks for domain/group/user access rights depending
    on the client's user and keypair information.

    :param autofill_user: When the *user_key* is not specified,
        automatically fills out the user data with the current
        user who is makeing the API request.
    :param user_key: The key used for storing user identification value
        in the keyword arguments.
    """

    def wrap(resolve_func):
        @functools.wraps(resolve_func)
        async def wrapped(
            root: Any,
            info: graphene.ResolveInfo,
            *args,
            **kwargs,
        ) -> Any:
            from .user import UserRole

            ctx: GraphQueryContext = info.context
            client_role = ctx.user["role"]
            if user_key == "access_key":
                client_user_id = ctx.access_key
            elif user_key == "email":
                client_user_id = ctx.user["email"]
            else:
                client_user_id = ctx.user["uuid"]
            client_domain = ctx.user["domain_name"]
            domain_name = kwargs.get("domain_name", None)
            group_id = kwargs.get("group_id", None) or kwargs.get("project_id", None)
            user_id = kwargs.get(user_key, None)
            if client_role == UserRole.SUPERADMIN:
                if autofill_user:
                    if user_id is None:
                        user_id = client_user_id
            elif client_role == UserRole.ADMIN:
                if domain_name is not None and domain_name != client_domain:
                    raise GenericForbidden
                domain_name = client_domain
                if group_id is not None:
                    # TODO: check if the group is a member of the domain
                    pass
                if autofill_user:
                    if user_id is None:
                        user_id = client_user_id
            elif client_role == UserRole.USER:
                if domain_name is not None and domain_name != client_domain:
                    raise GenericForbidden
                domain_name = client_domain
                if group_id is not None:
                    # TODO: check if the group is a member of the domain
                    # TODO: check if the client is a member of the group
                    pass
                if user_id is not None and user_id != client_user_id:
                    raise GenericForbidden
                user_id = client_user_id
            else:
                raise InvalidAPIParameters("Unknown client role")
            kwargs["domain_name"] = domain_name
            if group_id is not None:
                kwargs["group_id"] = group_id
            if kwargs.get("project", None) is not None:
                kwargs["project"] = group_id
            kwargs[user_key] = user_id
            return await resolve_func(root, info, *args, **kwargs)

        return wrapped

    return wrap


def privileged_mutation(required_role, target_func=None):
    def wrap(func):
        @functools.wraps(func)
        async def wrapped(cls, root, info: graphene.ResolveInfo, *args, **kwargs) -> Any:
            from .group import groups  # , association_groups_users
            from .user import UserRole

            ctx: GraphQueryContext = info.context
            permitted = False
            if required_role == UserRole.SUPERADMIN:
                if ctx.user["role"] == required_role:
                    permitted = True
            elif required_role == UserRole.ADMIN:
                if ctx.user["role"] == UserRole.SUPERADMIN:
                    permitted = True
                elif ctx.user["role"] == UserRole.USER:
                    permitted = False
                else:
                    if target_func is None:
                        return cls(False, "misconfigured privileged mutation: no target_func", None)
                    target_domain, target_group = target_func(*args, **kwargs)
                    if target_domain is None and target_group is None:
                        return cls(
                            False,
                            "misconfigured privileged mutation: "
                            "both target_domain and target_group missing",
                            None,
                        )
                    permit_chains = []
                    if target_domain is not None:
                        if ctx.user["domain_name"] == target_domain:
                            permit_chains.append(True)
                    if target_group is not None:
                        async with ctx.db.begin() as conn:
                            # check if the group is part of the requester's domain.
                            query = groups.select().where(
                                (groups.c.id == target_group)
                                & (groups.c.domain_name == ctx.user["domain_name"]),
                            )
                            result = await conn.execute(query)
                            if result.rowcount > 0:
                                permit_chains.append(True)
                            # TODO: check the group permission if implemented
                            # query = (
                            #     association_groups_users.select()
                            #     .where(association_groups_users.c.group_id == target_group)
                            # )
                            # result = await conn.execute(query)
                            # if result.rowcount > 0:
                            #     permit_chains.append(True)
                    permitted = all(permit_chains) if permit_chains else False
            elif required_role == UserRole.USER:
                permitted = True
            # assuming that mutation result objects has 2 or 3 fields:
            # success(bool), message(str) - usually for delete mutations
            # success(bool), message(str), item(object)
            if permitted:
                return await func(cls, root, info, *args, **kwargs)
            return cls(False, f"no permission to execute {info.path[0]}")

        return wrapped

    return wrap


ResultType = TypeVar("ResultType", bound=graphene.ObjectType)
ItemType = TypeVar("ItemType", bound=graphene.ObjectType)


async def gql_mutation_wrapper(
    result_cls: type[ResultType], _do_mutate: Callable[[], Awaitable[ResultType]]
) -> ResultType:
    try:
        return await execute_with_retry(_do_mutate)
    except sa.exc.IntegrityError as e:
        log.warning("gql_mutation_wrapper(): integrity error ({})", repr(e))
        return result_cls(False, f"integrity error: {e}")
    except sa.exc.StatementError as e:
        log.warning(
            "gql_mutation_wrapper(): statement error ({})\n{}", repr(e), e.statement or "(unknown)"
        )
        orig_exc = e.orig
        return result_cls(False, str(orig_exc), None)
    except (asyncio.CancelledError, asyncio.TimeoutError):
        raise
    except Exception as e:
        log.exception("gql_mutation_wrapper(): other error")
        return result_cls(False, f"unexpected error: {e}")


async def simple_db_mutate(
    result_cls: type[ResultType],
    graph_ctx: GraphQueryContext,
    mutation_query: sa.sql.Update | sa.sql.Insert | Callable[[], sa.sql.Update | sa.sql.Insert],
    *,
    pre_func: Callable[[SAConnection], Awaitable[None]] | None = None,
    post_func: Callable[[SAConnection, Result], Awaitable[None]] | None = None,
) -> ResultType:
    """
    Performs a database mutation based on the given
    :class:`sqlalchemy.sql.Update` or :class:`sqlalchemy.sql.Insert` query,
    and return the wrapped result as the GraphQL object type given as **result_cls**.
    **result_cls** should have two initialization arguments: success (bool)
    and message (str).

    See details about the arguments in :func:`simple_db_mutate_returning_item`.
    """

    async def _do_mutate() -> ResultType:
        async with graph_ctx.db.begin() as conn:
            if pre_func:
                await pre_func(conn)
            _query = mutation_query() if callable(mutation_query) else mutation_query
            result = await conn.execute(_query)
            if post_func:
                await post_func(conn, result)
        if result.rowcount > 0:
            return result_cls(True, "success")
        else:
            return result_cls(False, f"no matching {result_cls.__name__.lower()}")

    return await gql_mutation_wrapper(result_cls, _do_mutate)


async def simple_db_mutate_returning_item(
    result_cls: type[ResultType],
    graph_ctx: GraphQueryContext,
    mutation_query: sa.sql.Update | sa.sql.Insert | Callable[[], sa.sql.Update | sa.sql.Insert],
    *,
    item_cls: type[ItemType],
    pre_func: Callable[[SAConnection], Awaitable[None]] | None = None,
    post_func: Callable[[SAConnection, Result], Awaitable[Row]] | None = None,
) -> ResultType:
    """
    Performs a database mutation based on the given
    :class:`sqlalchemy.sql.Update` or :class:`sqlalchemy.sql.Insert` query,
    and return the wrapped result as the GraphQL object type given as **result_cls**
    and the inserted/updated row wrapped as its 3rd argument in **item_cls**.

    If mutation_query uses external variable updated by pre_func, you should wrap the query
    with lambda so that its parameters are re-evaluated when the transaction is retried.

    :param result_cls: The GraphQL Object Type used to wrap the result.
        It should have two initialization arguments: success (bool),
        message (str), and the item (ItemType).
    :param graph_ctx: The common context that provides the reference to the database engine
        and other stuffs required to resolve the GraphQL query.
    :param mutation_query: A SQLAlchemy query object.
    :param item_cls: The GraphQL Object Type used to wrap the returned row from the mutation query.
    :param pre_func: An extra function that is executed before the mutation query, where the caller
        may perform additional database queries.
    :param post_func: An extra function that is executed after the mutation query, where the caller
        may perform additional database queries.  Note that it **MUST return the returned row
        from the given mutation result**, because the result object could be fetched only one
        time due to its cursor-like nature.
    """

    async def _do_mutate() -> ResultType:
        async with graph_ctx.db.begin() as conn:
            if pre_func:
                await pre_func(conn)
            _query = mutation_query() if callable(mutation_query) else mutation_query
            _query = _query.returning(_query.table)
            result = await conn.execute(_query)
            if post_func:
                row = await post_func(conn, result)
            else:
                row = result.first()
            if result.rowcount > 0:
                return result_cls(True, "success", item_cls.from_row(graph_ctx, row))
            else:
                return result_cls(False, f"no matching {result_cls.__name__.lower()}", None)

    return await gql_mutation_wrapper(result_cls, _do_mutate)


def set_if_set(
    src: Any,
    target: MutableMapping[str, Any],
    name: str,
    *,
    clean_func=None,
    target_key: Optional[str] = None,
) -> None:
    """
    Set the target dict with only non-undefined keys and their values
    from a Graphene's input object.
    (server-side function)
    """
    match src:
        case Mapping():
            v = src.get(name, Undefined)
        case _:
            v = getattr(src, name)
    # NOTE: unset optional fields are passed as graphql.Undefined.
    if v is not Undefined:
        if callable(clean_func):
            target[target_key or name] = clean_func(v)
        else:
            target[target_key or name] = v


def orm_set_if_set(
    src: object,
    target: MutableMapping[str, Any],
    name: str,
    *,
    clean_func=None,
    target_key: Optional[str] = None,
) -> None:
    """
    Set the target ORM row object with only non-undefined keys and their values
    from a Graphene's input object.
    (server-side function)
    """
    v = getattr(src, name)
    # NOTE: unset optional fields are passed as graphql.Undefined.
    if v is not Undefined:
        if callable(clean_func):
            setattr(target, target_key or name, clean_func(v))
        else:
            setattr(target, target_key or name, v)


def filter_gql_undefined[T](val: T, *, default_value: Optional[T] = None) -> Optional[T]:
    if val is Undefined:
        return default_value
    else:
        return val


async def populate_fixture(
    engine: SAEngine,
    fixture_data: Mapping[str, str | Sequence[dict[str, Any]]],
) -> None:
    op_mode = FixtureOpModes(cast(str, fixture_data.get("__mode", "insert")))
    for table_name, rows in fixture_data.items():
        if table_name.startswith("__"):
            # skip reserved names like "__mode"
            continue
        if isinstance(rows, str):
            raise DataTransformationFailed(
                f"Invalid fixture data for table {table_name}: expected sequence, got string"
            )

        table: sa.Table = metadata.tables.get(table_name)

        if not isinstance(table, sa.Table):
            raise DataTransformationFailed(f"Table {table_name} not found in metadata")
        if not rows:
            return
        log.debug("Loading the fixture table {0} (mode:{1})", table_name, op_mode.name)
        from .user import PasswordColumn

        async with engine.begin() as conn:
            # Apply typedecorator manually for required columns
            for col in table.columns:
                if isinstance(col.type, sa.sql.sqltypes.DateTime):
                    for row in rows:
                        if col.name in row:
                            if row[col.name] is not None:
                                row[col.name] = isoparse(row[col.name])
                            else:
                                row[col.name] = None
                if isinstance(col.type, EnumType):
                    for row in rows:
                        if col.name in row:
                            row[col.name] = col.type._enum_cls[row[col.name]]
                elif isinstance(col.type, (StrEnumType, EnumValueType)):
                    for row in rows:
                        if col.name in row:
                            row[col.name] = col.type._enum_cls(row[col.name])
                elif isinstance(col.type, (StructuredJSONObjectColumn)):
                    for row in rows:
                        if col.name in row:
                            row[col.name] = col.type._schema.from_json(row[col.name])
                elif isinstance(col.type, (StructuredJSONObjectListColumn)):
                    for row in rows:
                        if col.name in row and row[col.name] is not None:
                            row[col.name] = [
                                item
                                if isinstance(item, col.type._schema)
                                else col.type._schema.from_json(item)
                                for item in row[col.name]
                            ]
                elif isinstance(col.type, PasswordColumn):
                    for row in rows:
                        if col.name in row and row[col.name] is not None:
                            # Convert raw password string to PasswordInfo
                            # Using default algorithm and parameters for fixtures
                            row[col.name] = PasswordInfo(
                                password=row[col.name],
                                algorithm=PasswordHashAlgorithm.PBKDF2_SHA256,
                                rounds=600_000,
                                salt_size=32,
                            )

            match op_mode:
                case FixtureOpModes.INSERT:
                    stmt = sa.dialects.postgresql.insert(table, rows).on_conflict_do_nothing()
                    await conn.execute(stmt)
                case FixtureOpModes.UPDATE:
                    stmt = sa.update(table)
                    pkcols = []
                    for pkidx, pkcol in enumerate(table.primary_key):
                        stmt = stmt.where(pkcol == sa.bindparam(f"_pk_{pkidx}"))
                        pkcols.append(pkcol)
                    update_data = []
                    # Extract the data column names from the FIRST row
                    # (Therefore a fixture dataset for a single table in the udpate mode should
                    # have consistent set of attributes!)
                    try:
                        datacols: list[sa.Column] = [
                            getattr(table.columns, name)
                            for name in set(rows[0].keys()) - {pkcol.name for pkcol in pkcols}
                        ]
                    except AttributeError as e:
                        raise ValueError(
                            f"fixture for table {table_name!r} has an invalid column name: "
                            f"{e.args[0]!r}"
                        )
                    stmt = stmt.values({
                        datacol.name: sa.bindparam(datacol.name) for datacol in datacols
                    })
                    for row in rows:
                        update_row = {}
                        for pkidx, pkcol in enumerate(pkcols):
                            try:
                                update_row[f"_pk_{pkidx}"] = row[pkcol.name]
                            except KeyError:
                                raise ValueError(
                                    f"fixture for table {table_name!r} has a missing primary key column for update"
                                    f"query: {pkcol.name!r}"
                                )
                        for datacol in datacols:
                            try:
                                update_row[datacol.name] = row[datacol.name]
                            except KeyError:
                                raise ValueError(
                                    f"fixture for table {table_name!r} has a missing data column for update"
                                    f"query: {datacol.name!r}"
                                )
                        update_data.append(update_row)
                    await conn.execute(stmt, update_data)


class InferenceSessionError(graphene.ObjectType):
    class InferenceSessionErrorInfo(graphene.ObjectType):
        src = graphene.String(required=True)
        name = graphene.String(required=True)
        repr = graphene.String(required=True)

    session_id = graphene.UUID()

    errors = graphene.List(graphene.NonNull(InferenceSessionErrorInfo), required=True)


class AsyncPaginatedConnectionField(AsyncListConnectionField):
    def __init__(self, type, *args, **kwargs):
        kwargs.setdefault("filter", graphene.String())
        kwargs.setdefault("order", graphene.String())
        kwargs.setdefault("offset", graphene.Int())
        super().__init__(type, *args, **kwargs)


PaginatedConnectionField = AsyncPaginatedConnectionField


class ConnectionArgs(NamedTuple):
    cursor: str | None
    pagination_order: ConnectionPaginationOrder | None
    requested_page_size: int


def validate_connection_args(
    *,
    after: str | None = None,
    first: int | None = None,
    before: str | None = None,
    last: int | None = None,
) -> ConnectionArgs:
    """
    Validate arguments used for GraphQL relay connection, and determine pagination ordering, cursor and page size.
    It is not allowed to use arguments for forward pagination and arguments for backward pagination at the same time.
    """
    order: ConnectionPaginationOrder | None = None
    cursor: str | None = None
    requested_page_size: int | None = None

    if after is not None:
        order = ConnectionPaginationOrder.FORWARD
        cursor = after
    if first is not None:
        if first < 0:
            raise ValueError("Argument 'first' must be a non-negative integer.")
        order = ConnectionPaginationOrder.FORWARD
        requested_page_size = first

    if before is not None:
        if order is ConnectionPaginationOrder.FORWARD:
            raise ValueError(
                "Can only paginate with single direction, forwards or backwards. Please set only"
                " one of (after, first) or (before, last)."
            )
        order = ConnectionPaginationOrder.BACKWARD
        cursor = before
    if last is not None:
        if last < 0:
            raise ValueError("Argument 'last' must be a non-negative integer.")
        if order is ConnectionPaginationOrder.FORWARD:
            raise ValueError(
                "Can only paginate with single direction, forwards or backwards. Please set only"
                " one of (after, first) or (before, last)."
            )
        order = ConnectionPaginationOrder.BACKWARD
        requested_page_size = last

    if requested_page_size is None:
        requested_page_size = DEFAULT_PAGE_SIZE

    return ConnectionArgs(cursor, order, requested_page_size)


def _build_sql_stmt_from_connection_args(
    info: graphene.ResolveInfo,
    orm_class,
    id_column: sa.Column,
    filter_expr: FilterExprArg | None = None,
    order_expr: OrderExprArg | None = None,
    *,
    connection_args: ConnectionArgs,
) -> tuple[sa.sql.Select, sa.sql.Select, list[WhereClauseType]]:
    stmt = sa.select(orm_class)
    count_stmt = sa.select(sa.func.count()).select_from(orm_class)
    conditions: list[WhereClauseType] = []

    cursor_id, pagination_order, requested_page_size = connection_args

    ordering_item_list: list[OrderingItem] = []
    if order_expr is not None:
        parser = order_expr.parser
        ordering_item_list = parser.parse_order(orm_class, order_expr.expr)

    # Apply SQL order_by
    match pagination_order:
        case ConnectionPaginationOrder.FORWARD | None:
            # Default ordering by id column
            id_ordering_item = OrderingItem(id_column, OrderDirection.ASC)
            set_ordering = lambda col, direction: (
                col.asc() if direction == OrderDirection.ASC else col.desc()
            )
        case ConnectionPaginationOrder.BACKWARD:
            # Default ordering by id column
            id_ordering_item = OrderingItem(id_column, OrderDirection.DESC)
            set_ordering = lambda col, direction: (
                col.desc() if direction == OrderDirection.ASC else col.asc()
            )
    # id column should be applied last
    for col, direction in [*ordering_item_list, id_ordering_item]:
        stmt = stmt.order_by(set_ordering(col, direction))

    # Set cursor by comparing scalar values of subquery that queried by cursor id
    if cursor_id is not None:
        _, cursor_row_id = AsyncNode.resolve_global_id(info, cursor_id)

        def subq_to_condition(
            column_to_be_compared: InstrumentedAttribute,
            subquery: ScalarSelect,
            direction: OrderDirection,
        ) -> WhereClauseType:
            match pagination_order:
                case ConnectionPaginationOrder.FORWARD | None:
                    if direction == OrderDirection.ASC:
                        cond = column_to_be_compared > subquery
                    else:
                        cond = column_to_be_compared < subquery

                    # Comparing ID field - The direction of inequality sign - is not effected by `direction` argument here
                    # because the ordering direction of ID field is always determined by `pagination_order` only.
                    condition_when_same_with_subq = (column_to_be_compared == subquery) & (
                        id_column > cursor_row_id
                    )
                case ConnectionPaginationOrder.BACKWARD:
                    if direction == OrderDirection.ASC:
                        cond = column_to_be_compared < subquery
                    else:
                        cond = column_to_be_compared > subquery
                    condition_when_same_with_subq = (column_to_be_compared == subquery) & (
                        id_column < cursor_row_id
                    )

            return cond | condition_when_same_with_subq

        for col, direction in ordering_item_list:
            subq = sa.select(col).where(id_column == cursor_row_id).scalar_subquery()
            conditions.append(subq_to_condition(col, subq, direction))

    if requested_page_size is not None:
        # Add 1 to determine has_next_page or has_previous_page
        stmt = stmt.limit(requested_page_size + 1)

    if filter_expr is not None:
        condition_parser = filter_expr.parser
        conditions.append(condition_parser.parse_filter(orm_class, filter_expr.expr))

    for cond in conditions:
        stmt = stmt.where(cond)
        count_stmt = count_stmt.where(cond)
    return stmt, count_stmt, conditions


def _build_sql_stmt_from_sql_arg(
    info: graphene.ResolveInfo,
    orm_class,
    id_column: sa.Column,
    filter_expr: FilterExprArg | None = None,
    order_expr: OrderExprArg | None = None,
    *,
    limit: int,
    offset: int | None = None,
) -> tuple[sa.sql.Select, sa.sql.Select, list[WhereClauseType]]:
    stmt = sa.select(orm_class)
    count_stmt = sa.select(sa.func.count()).select_from(orm_class)
    conditions: list[WhereClauseType] = []

    if order_expr is not None:
        parser = order_expr.parser
        stmt = parser.append_ordering(stmt, order_expr.expr)

    # default order_by id column
    stmt = stmt.order_by(id_column.asc())

    if filter_expr is not None:
        condition_parser = filter_expr.parser
        conditions.append(condition_parser.parse_filter(orm_class, filter_expr.expr))

    if limit is not None:
        stmt = stmt.limit(limit)

    if offset is not None:
        stmt = stmt.offset(offset)
    for cond in conditions:
        stmt = stmt.where(cond)
        count_stmt = count_stmt.where(cond)
    return stmt, count_stmt, conditions


class GraphQLConnectionSQLInfo(NamedTuple):
    sql_stmt: sa.sql.Select
    sql_count_stmt: sa.sql.Select
    sql_conditions: list[WhereClauseType]
    cursor: str | None
    pagination_order: ConnectionPaginationOrder | None
    requested_page_size: int


class FilterExprArg(NamedTuple):
    expr: str
    parser: QueryFilterParser


class OrderExprArg(NamedTuple):
    expr: str
    parser: QueryOrderParser


def generate_sql_info_for_gql_connection(
    info: graphene.ResolveInfo,
    orm_class,
    id_column: sa.Column,
    filter_expr: FilterExprArg | None = None,
    order_expr: OrderExprArg | None = None,
    offset: int | None = None,
    after: str | None = None,
    first: int | None = None,
    before: str | None = None,
    last: int | None = None,
) -> GraphQLConnectionSQLInfo:
    """
    Get GraphQL arguments and generate SQL query statement, cursor that points an id of a node, pagination order, and page size.
    If `offset` is None, return SQL query parsed from GraphQL Connection spec arguments.
    Else, return normally paginated SQL query and `first` is used as SQL limit.
    """

    if offset is None:
        connection_args = validate_connection_args(
            after=after, first=first, before=before, last=last
        )
        stmt, count_stmt, conditions = _build_sql_stmt_from_connection_args(
            info,
            orm_class,
            id_column,
            filter_expr,
            order_expr,
            connection_args=connection_args,
        )
        ret = GraphQLConnectionSQLInfo(
            stmt,
            count_stmt,
            conditions,
            connection_args.cursor,
            connection_args.pagination_order,
            connection_args.requested_page_size,
        )
    else:
        page_size = first if first is not None else DEFAULT_PAGE_SIZE
        stmt, count_stmt, conditions = _build_sql_stmt_from_sql_arg(
            info,
            orm_class,
            id_column,
            filter_expr,
            order_expr,
            limit=page_size,
            offset=offset,
        )
        ret = GraphQLConnectionSQLInfo(stmt, count_stmt, conditions, None, None, page_size)

    ctx: GraphQueryContext = info.context
    max_page_size = cast(Optional[int], ctx.config_provider.config.api.max_gql_connection_page_size)
    if max_page_size is not None and ret.requested_page_size > max_page_size:
        raise ValueError(
            f"Cannot fetch a page larger than {max_page_size}. "
            "Set 'first' or 'last' to a smaller integer."
        )
    return ret


class DecimalType(TypeDecorator, Decimal):
    """
    Database type adaptor for Decimal
    """

    impl = sa.VARCHAR
    cache_ok = True

    def process_bind_param(
        self,
        value: Optional[Decimal],
        dialect: Dialect,
    ) -> Optional[str]:
        return f"{value:f}" if value is not None else None

    def process_result_value(
        self,
        value: str,
        dialect: Dialect,
    ) -> Optional[Decimal]:
        return Decimal(value) if value is not None else None

    @property
    def python_type(self) -> type[Decimal]:
        return Decimal



from abc import abstractmethod
from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseAction


@dataclass
class NotificationAction(BaseAction):
    """Base action class for notification operations."""

    @override
    @classmethod
    def entity_type(cls) -> str:
        return "notification"

    @abstractmethod
    @override
    def entity_id(self) -> Optional[str]:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    @override
    def operation_type(cls) -> str:
        raise NotImplementedError



from typing import override

from ai.backend.common.metrics.metric import ActionMetricObserver
from ai.backend.manager.actions.action import BaseAction, BaseActionTriggerMeta, ProcessResult
from ai.backend.manager.actions.monitors.monitor import ActionMonitor


class PrometheusMonitor(ActionMonitor):
    _observer: ActionMetricObserver

    def __init__(self) -> None:
        self._observer = ActionMetricObserver.instance()

    @override
    async def prepare(self, action: BaseAction, meta: BaseActionTriggerMeta) -> None:
        return

    @override
    async def done(self, action: BaseAction, result: ProcessResult) -> None:
        self._observer.observe_action(
            entity_type=action.entity_type(),
            operation_type=action.operation_type(),
            status=result.meta.status,
            duration=result.meta.duration.total_seconds(),
            error_code=result.meta.error_code,
        )



"""Make artifact_revisions created_at, updated_at column nullable

Revision ID: 32764dcd0fad
Revises: 345551145801
Create Date: 2025-08-25 02:21:32.142541

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "32764dcd0fad"
down_revision = "345551145801"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "artifact_revisions",
        "created_at",
        existing_type=postgresql.TIMESTAMP(timezone=True),
        nullable=True,
        server_default=None,
        existing_server_default=sa.text("now()"),
    )
    op.alter_column(
        "artifact_revisions",
        "updated_at",
        existing_type=postgresql.TIMESTAMP(timezone=True),
        nullable=True,
        server_default=None,
        existing_server_default=sa.text("now()"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "artifact_revisions",
        "updated_at",
        existing_type=postgresql.TIMESTAMP(timezone=True),
        nullable=False,
        existing_server_default=sa.text("now()"),
    )
    op.alter_column(
        "artifact_revisions",
        "created_at",
        existing_type=postgresql.TIMESTAMP(timezone=True),
        nullable=False,
        existing_server_default=sa.text("now()"),
    )
    # ### end Alembic commands ###



import uuid
from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.storage_namespace.creator import StorageNamespaceCreator
from ai.backend.manager.data.storage_namespace.types import StorageNamespaceData
from ai.backend.manager.services.storage_namespace.actions.base import StorageNamespaceAction


@dataclass
class RegisterNamespaceAction(StorageNamespaceAction):
    creator: StorageNamespaceCreator

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "register"


@dataclass
class RegisterNamespaceActionResult(BaseActionResult):
    storage_id: uuid.UUID
    result: StorageNamespaceData

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.storage_id)



from collections.abc import Mapping
from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import StrEnum
from typing import Any

from ai.backend.common.types import AgentSelectionStrategy, SessionTypes


class SchedulerType(StrEnum):
    """Scheduler type for session scheduling."""

    FIFO = "fifo"
    LIFO = "lifo"
    DRF = "drf"


@dataclass
class ScalingGroupStatus:
    """Status information for a scaling group."""

    is_active: bool
    is_public: bool


@dataclass
class ScalingGroupMetadata:
    """Metadata for a scaling group."""

    description: str
    created_at: datetime


@dataclass
class ScalingGroupNetworkConfig:
    """Network configuration for a scaling group."""

    wsproxy_addr: str
    wsproxy_api_token: str
    use_host_network: bool


@dataclass
class ScalingGroupDriverConfig:
    """Driver configuration for a scaling group."""

    name: str
    options: Mapping[str, Any]


@dataclass
class ScalingGroupSchedulerOptions:
    """Scheduler options for a scaling group."""

    allowed_session_types: list[SessionTypes]
    pending_timeout: timedelta
    config: Mapping[str, Any]
    agent_selection_strategy: AgentSelectionStrategy
    agent_selector_config: Mapping[str, Any]
    enforce_spreading_endpoint_replica: bool
    allow_fractional_resource_fragmentation: bool
    route_cleanup_target_statuses: list[str]


@dataclass
class ScalingGroupSchedulerConfig:
    """Scheduler configuration for a scaling group."""

    name: SchedulerType
    options: ScalingGroupSchedulerOptions


@dataclass
class ScalingGroupData:
    name: str
    status: ScalingGroupStatus
    metadata: ScalingGroupMetadata
    network: ScalingGroupNetworkConfig
    driver: ScalingGroupDriverConfig
    scheduler: ScalingGroupSchedulerConfig


@dataclass
class ScalingGroupListResult:
    """Result of searching scaling groups."""

    items: list[ScalingGroupData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool



from .database import DatabaseHealthChecker
from .rpc import AgentRpcHealthChecker, RpcHealthCheckError

__all__ = [
    "AgentRpcHealthChecker",
    "DatabaseHealthChecker",
    "RpcHealthCheckError",
]



from typing import override

from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.processor import ActionProcessor
from ai.backend.manager.actions.types import AbstractProcessorPackage, ActionSpec
from ai.backend.manager.services.auth.actions.authorize import (
    AuthorizeAction,
    AuthorizeActionResult,
)
from ai.backend.manager.services.auth.actions.generate_ssh_keypair import (
    GenerateSSHKeypairAction,
    GenerateSSHKeypairActionResult,
)
from ai.backend.manager.services.auth.actions.get_role import GetRoleAction, GetRoleActionResult
from ai.backend.manager.services.auth.actions.get_ssh_keypair import (
    GetSSHKeypairAction,
    GetSSHKeypairActionResult,
)
from ai.backend.manager.services.auth.actions.signout import SignoutAction, SignoutActionResult
from ai.backend.manager.services.auth.actions.signup import SignupAction, SignupActionResult
from ai.backend.manager.services.auth.actions.update_full_name import (
    UpdateFullNameAction,
    UpdateFullNameActionResult,
)
from ai.backend.manager.services.auth.actions.update_password import (
    UpdatePasswordAction,
    UpdatePasswordActionResult,
)
from ai.backend.manager.services.auth.actions.update_password_no_auth import (
    UpdatePasswordNoAuthAction,
    UpdatePasswordNoAuthActionResult,
)
from ai.backend.manager.services.auth.actions.upload_ssh_keypair import (
    UploadSSHKeypairAction,
    UploadSSHKeypairActionResult,
)
from ai.backend.manager.services.auth.service import AuthService


class AuthProcessors(AbstractProcessorPackage):
    signout: ActionProcessor[SignoutAction, SignoutActionResult]
    update_full_name: ActionProcessor[UpdateFullNameAction, UpdateFullNameActionResult]
    get_ssh_keypair: ActionProcessor[GetSSHKeypairAction, GetSSHKeypairActionResult]
    generate_ssh_keypair: ActionProcessor[GenerateSSHKeypairAction, GenerateSSHKeypairActionResult]
    upload_ssh_keypair: ActionProcessor[UploadSSHKeypairAction, UploadSSHKeypairActionResult]
    get_role: ActionProcessor[GetRoleAction, GetRoleActionResult]
    authorize: ActionProcessor[AuthorizeAction, AuthorizeActionResult]
    signup: ActionProcessor[SignupAction, SignupActionResult]
    update_password: ActionProcessor[UpdatePasswordAction, UpdatePasswordActionResult]
    update_password_no_auth: ActionProcessor[
        UpdatePasswordNoAuthAction, UpdatePasswordNoAuthActionResult
    ]

    def __init__(self, service: AuthService, action_monitors: list[ActionMonitor]) -> None:
        self.signout = ActionProcessor(service.signout, action_monitors)
        self.update_full_name = ActionProcessor(service.update_full_name, action_monitors)
        self.get_ssh_keypair = ActionProcessor(service.get_ssh_keypair, action_monitors)
        self.generate_ssh_keypair = ActionProcessor(service.generate_ssh_keypair, action_monitors)
        self.upload_ssh_keypair = ActionProcessor(service.upload_ssh_keypair, action_monitors)
        self.get_role = ActionProcessor(service.get_role, action_monitors)
        self.authorize = ActionProcessor(service.authorize, action_monitors)
        self.signup = ActionProcessor(service.signup, action_monitors)
        self.update_password = ActionProcessor(service.update_password, action_monitors)
        self.update_password_no_auth = ActionProcessor(
            service.update_password_no_auth, action_monitors
        )

    @override
    def supported_actions(self) -> list[ActionSpec]:
        return [
            SignoutAction.spec(),
            UpdateFullNameAction.spec(),
            GetSSHKeypairAction.spec(),
            GenerateSSHKeypairAction.spec(),
            UploadSSHKeypairAction.spec(),
            GetRoleAction.spec(),
            AuthorizeAction.spec(),
            SignupAction.spec(),
            UpdatePasswordAction.spec(),
            UpdatePasswordNoAuthAction.spec(),
        ]



"""Route lifecycle handlers."""

from .base import RouteHandler
from .health_check import HealthCheckRouteHandler
from .provisioning import ProvisioningRouteHandler
from .route_eviction import RouteEvictionHandler
from .service_discovery_sync import ServiceDiscoverySyncHandler
from .terminating import TerminatingRouteHandler

__all__ = [
    "RouteHandler",
    "HealthCheckRouteHandler",
    "ProvisioningRouteHandler",
    "RouteEvictionHandler",
    "ServiceDiscoverySyncHandler",
    "TerminatingRouteHandler",
]



import uuid
from dataclasses import dataclass
from typing import Any, override

from ai.backend.manager.types import Creator, OptionalState

from .id import ObjectId
from .status import PermissionStatus
from .types import EntityType, OperationType


@dataclass
class ObjectPermissionCreateInput(Creator):
    role_id: uuid.UUID
    entity_type: EntityType
    entity_id: str
    operation: OperationType
    status: PermissionStatus = PermissionStatus.ACTIVE

    @override
    def fields_to_store(self) -> dict[str, Any]:
        return {
            "role_id": self.role_id,
            "entity_type": self.entity_type,
            "entity_id": self.entity_id,
            "operation": self.operation,
            "status": self.status,
        }


@dataclass
class ObjectPermissionCreateInputBeforeRoleCreation:
    entity_type: EntityType
    entity_id: str
    operation: OperationType
    status: PermissionStatus = PermissionStatus.ACTIVE

    def to_input(self, role_id: uuid.UUID) -> ObjectPermissionCreateInput:
        return ObjectPermissionCreateInput(
            role_id=role_id,
            entity_type=self.entity_type,
            entity_id=self.entity_id,
            operation=self.operation,
            status=self.status,
        )


@dataclass
class ObjectPermissionUpdater:
    id: uuid.UUID
    status: OptionalState[PermissionStatus]


@dataclass
class ObjectPermissionDeleteInput:
    id: uuid.UUID
    _status: PermissionStatus = PermissionStatus.DELETED


@dataclass
class ObjectPermissionData:
    id: uuid.UUID
    role_id: uuid.UUID
    object_id: ObjectId
    operation: OperationType



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Optional
from uuid import UUID

from ai.backend.common.types import BinarySize, ResourceSlot


@dataclass
class ResourcePresetData:
    id: UUID
    name: str
    resource_slots: ResourceSlot
    shared_memory: Optional[int]
    scaling_group_name: Optional[str]

    def to_cache(self) -> dict[str, Any]:
        """Serialize to cache-friendly format."""
        return {
            "id": str(self.id),
            "name": self.name,
            "resource_slots": self.resource_slots.to_json(),
            "shared_memory": str(self.shared_memory) if self.shared_memory is not None else None,
            "scaling_group_name": self.scaling_group_name,
        }

    @classmethod
    def from_cache(cls, data: dict[str, Any]) -> ResourcePresetData:
        """Deserialize from cache format."""
        return cls(
            id=UUID(data["id"]),
            name=data["name"],
            resource_slots=ResourceSlot.from_json(data["resource_slots"]),
            shared_memory=int(BinarySize.from_str(data["shared_memory"]))
            if data["shared_memory"] is not None
            else None,
            scaling_group_name=data["scaling_group_name"],
        )



"""Separate designated agent ids

Revision ID: 4db74efaf82e
Revises: fb01c44db0d6
Create Date: 2025-09-03 12:52:38.558284

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "4db74efaf82e"
down_revision = "fb01c44db0d6"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "sessions", sa.Column("designated_agent_ids", sa.ARRAY(sa.String()), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("sessions", "designated_agent_ids")
    # ### end Alembic commands ###



import logging
from typing import TYPE_CHECKING

from ai.backend.common.clients.valkey_client.valkey_schedule import ValkeyScheduleClient
from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.common.leader.tasks import EventTaskSpec
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.sokovan.deployment.coordinator import DeploymentCoordinator
from ai.backend.manager.sokovan.deployment.route.coordinator import RouteCoordinator
from ai.backend.manager.sokovan.scheduler.coordinator import ScheduleCoordinator
from ai.backend.manager.sokovan.scheduler.scheduler import Scheduler
from ai.backend.manager.sokovan.scheduling_controller import SchedulingController

if TYPE_CHECKING:
    from ai.backend.manager.types import DistributedLockFactory

log = BraceStyleAdapter(logging.getLogger(__name__))


class SokovanOrchestrator:
    """
    Orchestrator for Sokovan scheduler.
    Provides event handler interface for the coordinator.
    """

    _deployment_coordinator: DeploymentCoordinator
    _route_coordinator: RouteCoordinator
    _schedule_coordinator: ScheduleCoordinator
    _event_producer: EventProducer
    _lock_factory: "DistributedLockFactory"
    _scheduling_controller: SchedulingController

    def __init__(
        self,
        scheduler: Scheduler,
        event_producer: EventProducer,
        valkey_schedule: ValkeyScheduleClient,
        lock_factory: "DistributedLockFactory",
        scheduling_controller: SchedulingController,
        deployment_coordinator: DeploymentCoordinator,
        route_coordinator: RouteCoordinator,
    ) -> None:
        # Store injected coordinators
        self._deployment_coordinator = deployment_coordinator
        self._route_coordinator = route_coordinator

        # Initialize schedule coordinator
        self._schedule_coordinator = ScheduleCoordinator(
            valkey_schedule=valkey_schedule,
            scheduler=scheduler,
            scheduling_controller=scheduling_controller,
            event_producer=event_producer,
            lock_factory=lock_factory,
            config_provider=scheduler._config_provider,
        )

        self._event_producer = event_producer
        self._lock_factory = lock_factory
        self._scheduling_controller = scheduling_controller

    @property
    def coordinator(self) -> ScheduleCoordinator:
        """Get the schedule coordinator."""
        return self._schedule_coordinator

    @property
    def deployment_coordinator(self) -> DeploymentCoordinator:
        """Get the deployment coordinator."""
        return self._deployment_coordinator

    @property
    def route_coordinator(self) -> RouteCoordinator:
        """Get the route coordinator."""
        return self._route_coordinator

    def create_task_specs(self) -> list[EventTaskSpec]:
        """Create task specifications for leader-based scheduling."""
        specs: list[EventTaskSpec] = []

        # Collect task specs from all coordinators
        specs.extend(self._schedule_coordinator.create_task_specs())
        specs.extend(self._deployment_coordinator.create_task_specs())
        specs.extend(self._route_coordinator.create_task_specs())

        return specs



import uuid
from typing import Any, Optional

import sqlalchemy as sa
from sqlalchemy.ext.asyncio import AsyncSession as SASession
from sqlalchemy.orm.exc import NoResultFound

from ai.backend.common.clients.valkey_client.valkey_live.client import ValkeyLiveClient
from ai.backend.common.exception import BackendAIError
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience
from ai.backend.manager.data.model_serving.types import (
    EndpointAutoScalingRuleData,
    EndpointData,
    EndpointTokenData,
    RoutingData,
)
from ai.backend.manager.errors.common import ObjectNotFound
from ai.backend.manager.models.endpoint import (
    AutoScalingMetricComparator,
    AutoScalingMetricSource,
    EndpointAutoScalingRuleRow,
    EndpointLifecycle,
    EndpointRow,
    EndpointTokenRow,
)
from ai.backend.manager.models.routing import RouteStatus, RoutingRow
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

model_serving_repository_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(domain=DomainType.REPOSITORY, layer=LayerType.MODEL_SERVING_REPOSITORY)
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=10,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class AdminModelServingRepository:
    """
    Repository for admin-specific model serving operations that bypass ownership checks.
    This should only be used by superadmin users.
    """

    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    @model_serving_repository_resilience.apply()
    async def get_endpoint_by_id_force(self, endpoint_id: uuid.UUID) -> Optional[EndpointData]:
        """
        Get endpoint by ID without access control validation.
        Returns None if endpoint doesn't exist.
        """
        async with self._db.begin_readonly_session() as session:
            endpoint = await self._get_endpoint_by_id(
                session,
                endpoint_id,
                load_routes=True,
                load_session_owner=True,
                load_model=True,
                load_image=True,
            )
            if not endpoint:
                return None
            data = endpoint.to_data()
        return data

    @model_serving_repository_resilience.apply()
    async def update_endpoint_lifecycle_force(
        self,
        endpoint_id: uuid.UUID,
        lifecycle_stage: EndpointLifecycle,
        replicas: Optional[int] = None,
    ) -> bool:
        """
        Update endpoint lifecycle stage without access validation.
        Returns True if updated, False if not found.
        """
        async with self._db.begin_session() as session:
            endpoint = await self._get_endpoint_by_id(session, endpoint_id)
            if not endpoint:
                return False

            update_values: dict[str, Any] = {"lifecycle_stage": lifecycle_stage}
            if lifecycle_stage == EndpointLifecycle.DESTROYED:
                update_values["destroyed_at"] = sa.func.now()
            if replicas is not None:
                update_values["replicas"] = replicas

            query = (
                sa.update(EndpointRow).where(EndpointRow.id == endpoint_id).values(update_values)
            )
            await session.execute(query)
        return True

    @model_serving_repository_resilience.apply()
    async def clear_endpoint_errors_force(self, endpoint_id: uuid.UUID) -> bool:
        """
        Clear endpoint errors (failed routes and reset retry count) without access validation.
        Returns True if cleared, False if not found.
        """
        async with self._db.begin_session() as session:
            endpoint = await self._get_endpoint_by_id(session, endpoint_id)
            if not endpoint:
                return False

            # Delete failed routes
            query = sa.delete(RoutingRow).where(
                (RoutingRow.endpoint == endpoint_id)
                & (RoutingRow.status == RouteStatus.FAILED_TO_START)
            )
            await session.execute(query)

            # Reset retry count
            query = (
                sa.update(EndpointRow).values({"retries": 0}).where(EndpointRow.id == endpoint_id)
            )
            await session.execute(query)
        return True

    @model_serving_repository_resilience.apply()
    async def get_route_by_id_force(
        self, route_id: uuid.UUID, service_id: uuid.UUID
    ) -> Optional[RoutingData]:
        """
        Get route by ID without endpoint ownership validation.
        Returns None if route doesn't exist or doesn't belong to service.
        """
        async with self._db.begin_readonly_session() as session:
            route = await self._get_route_by_id(session, route_id, load_endpoint=True)
            if not route or route.endpoint != service_id:
                return None
            data = route.to_data()
        return data

    @model_serving_repository_resilience.apply()
    async def update_route_traffic_force(
        self,
        valkey_live: ValkeyLiveClient,
        route_id: uuid.UUID,
        service_id: uuid.UUID,
        traffic_ratio: float,
    ) -> Optional[EndpointData]:
        """
        Update route traffic ratio without access validation.
        Returns updated endpoint data if successful, None if not found.
        """
        async with self._db.begin_session() as session:
            route = await self._get_route_by_id(session, route_id, load_endpoint=True)
            if not route or route.endpoint != service_id:
                return None

            query = (
                sa.update(RoutingRow)
                .where(RoutingRow.id == route_id)
                .values({"traffic_ratio": traffic_ratio})
            )
            await session.execute(query)

            endpoint = await self._get_endpoint_by_id(session, service_id, load_routes=True)
            if endpoint is None:
                raise NoResultFound
            data = endpoint.to_data()

            await valkey_live.store_live_data(
                f"endpoint.{service_id}.session.{route.session}.traffic_ratio",
                str(traffic_ratio),
            )
        return data

    @model_serving_repository_resilience.apply()
    async def decrease_endpoint_replicas_force(self, service_id: uuid.UUID) -> bool:
        """
        Decrease endpoint replicas by 1 without access validation.
        Returns True if decreased, False if not found.
        """
        async with self._db.begin_session() as session:
            endpoint = await self._get_endpoint_by_id(session, service_id)
            if not endpoint:
                return False

            query = (
                sa.update(EndpointRow)
                .where(EndpointRow.id == service_id)
                .values({"replicas": endpoint.replicas - 1})
            )
            await session.execute(query)
        return True

    @model_serving_repository_resilience.apply()
    async def create_endpoint_token_force(
        self, token_row: EndpointTokenRow
    ) -> Optional[EndpointTokenData]:
        """
        Create endpoint token without access validation.
        Returns token data if created, None if endpoint not found.
        """
        async with self._db.begin_session() as session:
            endpoint = await self._get_endpoint_by_id(session, token_row.endpoint)
            if not endpoint:
                return None

            session.add(token_row)
            await session.commit()
            await session.refresh(token_row)
            data = token_row.to_dataclass()
        return data

    async def _get_endpoint_by_id(
        self,
        session: SASession,
        endpoint_id: uuid.UUID,
        load_routes: bool = False,
        load_session_owner: bool = False,
        load_model: bool = False,
        load_image: bool = False,
    ) -> Optional[EndpointRow]:
        """
        Private method to get endpoint by ID using an existing session.
        """
        try:
            return await EndpointRow.get(
                session,
                endpoint_id,
                load_routes=load_routes,
                load_session_owner=load_session_owner,
                load_model=load_model,
                load_image=load_image,
            )
        except NoResultFound:
            return None

    async def _get_route_by_id(
        self,
        session: SASession,
        route_id: uuid.UUID,
        load_endpoint: bool = False,
        load_session: bool = False,
    ) -> Optional[RoutingRow]:
        """
        Private method to get route by ID using an existing session.
        """
        try:
            return await RoutingRow.get(
                session, route_id, load_endpoint=load_endpoint, load_session=load_session
            )
        except NoResultFound:
            return None

    @model_serving_repository_resilience.apply()
    async def update_endpoint_replicas_force(self, endpoint_id: uuid.UUID, replicas: int) -> bool:
        """
        Update endpoint replicas without access validation.
        Returns True if updated, False if not found.
        """
        async with self._db.begin_session() as session:
            endpoint = await self._get_endpoint_by_id(session, endpoint_id)
            if not endpoint:
                return False

            query = (
                sa.update(EndpointRow)
                .where(EndpointRow.id == endpoint_id)
                .values({"replicas": replicas})
            )
            await session.execute(query)
        return True

    @model_serving_repository_resilience.apply()
    async def get_auto_scaling_rule_by_id_force(
        self, rule_id: uuid.UUID
    ) -> Optional[EndpointAutoScalingRuleData]:
        """
        Get auto scaling rule by ID without access validation.
        Returns None if rule doesn't exist.
        """
        async with self._db.begin_readonly_session() as session:
            try:
                row = await EndpointAutoScalingRuleRow.get(session, rule_id, load_endpoint=True)
            except ObjectNotFound:
                return None
            return row.to_data()

    @model_serving_repository_resilience.apply()
    async def create_auto_scaling_rule_force(
        self,
        endpoint_id: uuid.UUID,
        metric_source: AutoScalingMetricSource,
        metric_name: str,
        threshold: Any,
        comparator: AutoScalingMetricComparator,
        step_size: int,
        cooldown_seconds: int = 300,
        min_replicas: Optional[int] = None,
        max_replicas: Optional[int] = None,
    ) -> Optional[EndpointAutoScalingRuleData]:
        """
        Create auto scaling rule without access validation.
        Returns the created rule if successful, None if endpoint not found.
        """
        async with self._db.begin_session() as session:
            endpoint = await self._get_endpoint_by_id(session, endpoint_id)
            if not endpoint:
                return None

            if endpoint.lifecycle_stage in EndpointLifecycle.inactive_states():
                return None

            rule = await endpoint.create_auto_scaling_rule(
                session,
                metric_source,
                metric_name,
                threshold,
                comparator,
                step_size,
                cooldown_seconds=cooldown_seconds,
                min_replicas=min_replicas,
                max_replicas=max_replicas,
            )
            return rule.to_data()

    @model_serving_repository_resilience.apply()
    async def update_auto_scaling_rule_force(
        self, rule_id: uuid.UUID, fields_to_update: dict[str, Any]
    ) -> Optional[EndpointAutoScalingRuleData]:
        """
        Update auto scaling rule without access validation.
        Returns the updated rule if successful, None if not found.
        """
        async with self._db.begin_session() as session:
            try:
                rule = await EndpointAutoScalingRuleRow.get(session, rule_id, load_endpoint=True)
                if not rule:
                    return None

                if rule.endpoint_row.lifecycle_stage in EndpointLifecycle.inactive_states():
                    return None

                for key, value in fields_to_update.items():
                    setattr(rule, key, value)

                return rule.to_data()
            except ObjectNotFound:
                return None

    @model_serving_repository_resilience.apply()
    async def delete_auto_scaling_rule_force(self, rule_id: uuid.UUID) -> bool:
        """
        Delete auto scaling rule without access validation.
        Returns True if deleted, False if not found.
        """
        async with self._db.begin_session() as session:
            try:
                rule = await EndpointAutoScalingRuleRow.get(session, rule_id, load_endpoint=True)
                if not rule:
                    return False

                await session.delete(rule)
                return True
            except NoResultFound:
                return False



from __future__ import annotations

import strawberry
from strawberry import Info
from strawberry.scalars import JSON

from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.services.agent.actions.get_total_resources import GetTotalResourcesAction


@strawberry.field(description="Added in 25.15.0")
async def agent_stats(info: Info[StrawberryGQLContext]) -> AgentStats:
    result = await info.context.processors.agent.get_total_resources.wait_for_complete(
        GetTotalResourcesAction()
    )

    return AgentStats(
        total_resource=AgentResource(
            free=result.total_resources.total_free_slots.to_json(),
            used=result.total_resources.total_used_slots.to_json(),
            capacity=result.total_resources.total_capacity_slots.to_json(),
        )
    )


@strawberry.type(description="Added in 25.15.0")
class AgentResource:
    free: JSON
    used: JSON
    capacity: JSON


@strawberry.type(description="Added in 25.15.0")
class AgentStats:
    total_resource: AgentResource = strawberry.field(description="Added in 25.15.0")



from collections.abc import Mapping
from typing import Any, override

from ai.backend.common.etcd import AsyncEtcd
from ai.backend.manager.config.loader.types import AbstractConfigLoader


class EtcdCommonConfigLoader(AbstractConfigLoader):
    _etcd: AsyncEtcd

    def __init__(self, etcd: AsyncEtcd):
        self._etcd = etcd

    @override
    async def load(self) -> Mapping[str, Any]:
        raw_cfg = await self._etcd.get_prefix("ai/backend/config/common")
        return raw_cfg


class EtcdManagerConfigLoader(AbstractConfigLoader):
    _etcd: AsyncEtcd

    def __init__(self, etcd: AsyncEtcd):
        self._etcd = etcd

    @override
    async def load(self) -> Mapping[str, Any]:
        raw_cfg = await self._etcd.get_prefix("ai/backend/config/manager")
        return raw_cfg



"""GraphQL query resolvers for scaling group system."""

from __future__ import annotations

from functools import lru_cache
from typing import Optional

import strawberry
from strawberry import ID, Info
from strawberry.relay import Connection, Edge

from ai.backend.manager.api.gql.adapter import PaginationOptions, PaginationSpec
from ai.backend.manager.api.gql.base import encode_cursor
from ai.backend.manager.repositories.scaling_group.options import (
    ScalingGroupConditions,
    ScalingGroupOrders,
)
from ai.backend.manager.services.scaling_group.actions.list_scaling_groups import (
    SearchScalingGroupsAction,
)

from ..types import StrawberryGQLContext
from .types import (
    ScalingGroupFilterGQL,
    ScalingGroupOrderByGQL,
    ScalingGroupV2GQL,
)

# Pagination specs


@lru_cache(maxsize=1)
def _get_scaling_group_pagination_spec() -> PaginationSpec:
    return PaginationSpec(
        forward_order=ScalingGroupOrders.created_at(ascending=False),
        backward_order=ScalingGroupOrders.created_at(ascending=True),
        forward_condition_factory=ScalingGroupConditions.by_cursor_forward,
        backward_condition_factory=ScalingGroupConditions.by_cursor_backward,
    )


# Connection types

ScalingGroupV2Edge = Edge[ScalingGroupV2GQL]


@strawberry.type(description="Added in 25.18.0. Scaling group connection")
class ScalingGroupV2Connection(Connection[ScalingGroupV2GQL]):
    count: int

    def __init__(self, *args, count: int, **kwargs):
        super().__init__(*args, **kwargs)
        self.count = count


# Query fields


@strawberry.field(description="Added in 25.18.0. List scaling groups for a specific project")
async def scaling_groups_v2(
    info: Info[StrawberryGQLContext],
    project: ID,
    filter: Optional[ScalingGroupFilterGQL] = None,
    order_by: Optional[list[ScalingGroupOrderByGQL]] = None,
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> ScalingGroupV2Connection:
    processors = info.context.processors

    # Build querier from filter, order_by, and pagination using adapter
    querier = info.context.gql_adapter.build_querier(
        PaginationOptions(
            first=first,
            after=after,
            last=last,
            before=before,
            limit=limit,
            offset=offset,
        ),
        _get_scaling_group_pagination_spec(),
        filter=filter,
        order_by=order_by,
    )

    # Apply project filtering as a required condition
    project_condition = ScalingGroupConditions.by_project(project)
    querier.conditions.append(project_condition)

    action_result = await processors.scaling_group.search_scaling_groups.wait_for_complete(
        SearchScalingGroupsAction(querier=querier)
    )

    nodes = [ScalingGroupV2GQL.from_dataclass(data) for data in action_result.scaling_groups]

    edges = [ScalingGroupV2Edge(node=node, cursor=encode_cursor(node.id)) for node in nodes]

    return ScalingGroupV2Connection(
        edges=edges,
        page_info=strawberry.relay.PageInfo(
            has_next_page=action_result.has_next_page,
            has_previous_page=action_result.has_previous_page,
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if edges else None,
        ),
        count=action_result.total_count,
    )


@strawberry.field(description="Added in 25.18.0. List all scaling groups")
async def all_scaling_groups_v2(
    info: Info[StrawberryGQLContext],
    filter: Optional[ScalingGroupFilterGQL] = None,
    order_by: Optional[list[ScalingGroupOrderByGQL]] = None,
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> ScalingGroupV2Connection:
    processors = info.context.processors

    # Build querier from filter, order_by, and pagination using adapter
    querier = info.context.gql_adapter.build_querier(
        PaginationOptions(
            first=first,
            after=after,
            last=last,
            before=before,
            limit=limit,
            offset=offset,
        ),
        _get_scaling_group_pagination_spec(),
        filter=filter,
        order_by=order_by,
    )

    action_result = await processors.scaling_group.search_scaling_groups.wait_for_complete(
        SearchScalingGroupsAction(querier=querier)
    )

    nodes = [ScalingGroupV2GQL.from_dataclass(data) for data in action_result.scaling_groups]

    edges = [ScalingGroupV2Edge(node=node, cursor=encode_cursor(node.id)) for node in nodes]

    return ScalingGroupV2Connection(
        edges=edges,
        page_info=strawberry.relay.PageInfo(
            has_next_page=action_result.has_next_page,
            has_previous_page=action_result.has_previous_page,
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if edges else None,
        ),
        count=action_result.total_count,
    )



"""GraphQL notification module."""

from .resolver import (
    create_notification_channel,
    create_notification_rule,
    delete_notification_channel,
    delete_notification_rule,
    notification_channel,
    notification_channels,
    notification_rule,
    notification_rule_types,
    notification_rules,
    update_notification_channel,
    update_notification_rule,
    validate_notification_channel,
    validate_notification_rule,
)
from .types import (
    CreateNotificationChannelInput,
    CreateNotificationChannelPayload,
    CreateNotificationRuleInput,
    CreateNotificationRulePayload,
    DeleteNotificationChannelInput,
    DeleteNotificationChannelPayload,
    DeleteNotificationRuleInput,
    DeleteNotificationRulePayload,
    NotificationChannel,
    NotificationChannelFilter,
    NotificationChannelOrderBy,
    NotificationRule,
    NotificationRuleFilter,
    NotificationRuleOrderBy,
    UpdateNotificationChannelInput,
    UpdateNotificationChannelPayload,
    UpdateNotificationRuleInput,
    UpdateNotificationRulePayload,
    ValidateNotificationChannelInput,
    ValidateNotificationChannelPayload,
    ValidateNotificationRuleInput,
    ValidateNotificationRulePayload,
)

__all__ = (
    # Types
    "NotificationChannel",
    "NotificationRule",
    "NotificationChannelFilter",
    "NotificationChannelOrderBy",
    "NotificationRuleFilter",
    "NotificationRuleOrderBy",
    "CreateNotificationChannelInput",
    "CreateNotificationChannelPayload",
    "UpdateNotificationChannelInput",
    "UpdateNotificationChannelPayload",
    "DeleteNotificationChannelInput",
    "DeleteNotificationChannelPayload",
    "CreateNotificationRuleInput",
    "CreateNotificationRulePayload",
    "UpdateNotificationRuleInput",
    "UpdateNotificationRulePayload",
    "DeleteNotificationRuleInput",
    "DeleteNotificationRulePayload",
    "ValidateNotificationChannelInput",
    "ValidateNotificationChannelPayload",
    "ValidateNotificationRuleInput",
    "ValidateNotificationRulePayload",
    # Resolvers
    "notification_channel",
    "notification_channels",
    "notification_rule",
    "notification_rules",
    "notification_rule_types",
    "create_notification_channel",
    "update_notification_channel",
    "delete_notification_channel",
    "create_notification_rule",
    "update_notification_rule",
    "delete_notification_rule",
    "validate_notification_channel",
    "validate_notification_rule",
)



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass
from pathlib import Path

from ai.backend.common.dependencies import NonMonitorableDependencyProvider
from ai.backend.logging.types import LogLevel
from ai.backend.manager.config.bootstrap import BootstrapConfig


@dataclass
class BootstrapConfigInput:
    """Input required for bootstrap configuration loading.

    Contains only the essential parameters needed to load the bootstrap config.
    """

    config_path: Path
    log_level: LogLevel = LogLevel.NOTSET


class BootstrapConfigDependency(
    NonMonitorableDependencyProvider[BootstrapConfigInput, BootstrapConfig]
):
    """Provides BootstrapConfig lifecycle management.

    Loads bootstrap configuration from file with log level overrides.
    """

    @property
    def stage_name(self) -> str:
        return "bootstrap-config"

    @asynccontextmanager
    async def provide(self, setup_input: BootstrapConfigInput) -> AsyncIterator[BootstrapConfig]:
        """Load and provide bootstrap configuration.

        Args:
            setup_input: Input containing config path and log level

        Yields:
            Loaded BootstrapConfig
        """
        config = await BootstrapConfig.load_from_file(
            setup_input.config_path,
            setup_input.log_level,
        )
        yield config
